//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//


#include <vector>
#include <tuple>
#include <functional>
#include "mystring.h"

// デバッグ文字列を返す
// プリミティブ型
template <class T>
std::string
ToDebugString(T x) {
	return std::to_string(x);
}

// デバッグ文字列を返す
// bool 型
static std::string
ToDebugString(bool x)
{
	return x ? "true" : "false";
}

// デバッグ文字列を返す
// string 型
static std::string
ToDebugString(const std::string& x)
{
	std::string rv { '"' };
	// TODO: escape
	rv += x;
	rv += '"';
	return rv;
}

static std::string
ToDebugString(const char *x)
{
	return ToDebugString(std::string { x });
}

// Tuple with ToDebugString

// 前方宣言
template <class... Args>
class TestTuple;

// デバッグ文字列を返す
// TestTuple 型
template <class... Args>
std::string
ToDebugString(const TestTuple<Args...>& x)
{
	return x.ToDebugString();
}

// オレオレタプル
template <class... Args>
class TestTuple : public std::tuple<Args...>
{
	// 継承コンストラクタ宣言
	using std::tuple<Args...>::tuple;

 public:
	// デバッグ文字列を返す
	std::string ToDebugString() const {
		return ToD(std::index_sequence_for<Args...>{});
	}

	// std::get のメンバ関数化
	// 型が決まらないといけないので [] には出来ない。
	template <size_t i>
	auto get() const {
		return std::get<i>(*this);
	}

 private:
	// このへん参照
	// https://theolizer.com/cpp-school2/cpp-school2-21/

	template <size_t... indices>
	std::string ToD(std::index_sequence<indices...>) const {
		// get<> にはテンプレート引数を渡さないといけないので
		// 配列要素数がコンパイル時展開になるようにする。
		// これでタプルの各要素が型ごとに異なる ToDebugString()
		// の実体関数を呼び出して、結果の文字列が配列に入っている
		// 状態になる、ようだ。
		std::string a[] = {
			(::ToDebugString(get<indices>()))...
		};

		// a[] をよしなに連結すれば良い
		std::string rv { '{' };
		bool first = true;
		for (const auto& s : a) {
			if (first) {
				first = false;
			} else {
				rv += ',';
			}
			rv += s;
		}
		rv += '}';
		return rv;
	}
};

// テストデータ用テーブル
template <class... Args>
class TestTable : public std::vector<TestTuple<Args...> >
{
	// 継承コンストラクタ宣言
	using std::vector<TestTuple<Args... > >::vector;
};


// Test

class TestClass
{
 public:
	virtual ~TestClass() {}

	template <class T, class R>
	void Fail(T expectobj, R real) {
		printf("Failed at: %s but %s\n",
			ToDebugString(expectobj).c_str(),
			ToDebugString(real).c_str());
		exit(1);
	}

	void Test() {
		Init();

		for (auto&& t : testlist) {
			printf("Testing: %s ", t.name.c_str());
			t.func();
			printf("Ok\n");
		}
	}

 protected:
	// 継承して、TESTDEF マクロを使って testlist を作ってください。
	virtual void Init() = 0;

	// テスト関数のリスト
	class Entry {
	 public:
		std::string name;
		std::function<void(void)> func;
	};
	std::vector<Entry> testlist;

#define TESTDEF(x) testlist.push_back({ #x, [&]() { x(); } })


	template <class T, class TWhere>
	void xp_eq(T expect, T actual, const TWhere& where)
	{
		// eq なので == で比較。演算子オーバーロード対策。
		if (expect == actual) {
			return;
		}

		Fail(where, actual);
	}
};

class Test_MyString : public TestClass
{
 protected:
	void Init() override {
		// TESTDEF マクロに、テスト実行関数の識別子を渡します。
		TESTDEF(startwith_ignorecase_1);
		TESTDEF(string_ltrim_1);
	}

	void startwith_ignorecase_1() {
		TestTable<std::string, std::string, bool> table = {
			{ "ABCDEF", "A", true },
			{ "ABCDEF", "B", false },
			{ "ABCDEF", "AB", true },
			{ "ABCDEF", "BC", false },
			{ "ABCDEF", "a", true },
			{ "ABCDEF", "b", false },
			{ "ABCDEF", "ab", true },
			{ "ABCDEF", "bc", false },
			{ "aBCDEF", "a", true },
			{ "aBCDEF", "b", false },
			{ "aBCDEF", "ab", true },
			{ "aBCDEF", "ABcd", true },
			{ "", "A", false },
			{ "", "", true },	// ??? これはまだわかる
			{ "A", "", true },	// ??? 概念的には true だが果たして?
		};

		for (auto t : table) {
			auto arg0 = t.get<0>();
			auto arg1 = t.get<1>();
			auto expect = t.get<2>();

			xp_eq(expect, starts_with_ignorecase(arg0, arg1), t);
		}
	}

	void string_ltrim_1()
	{
		Fail("NOT IMPL", "");
	}
};

int
main(int ac, char *av[])
{
	printf("compiled by ");
#if defined(__clang_version__)
	printf("clang %s", __clang_version__);
#elif defined(__GNUC__) && defined(__VERSION__)
	printf("gcc %s", __VERSION__);
#else
	printf("unknown compiler");
#endif
	printf("\n");

	Test_MyString{}.Test();
	return 0;
}
