//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "m88100acc.h"
#include <cfenv>
#include <cmath>

// 命令のフィールド取り出し
#define FLD_D	m88100opf_D(opX)
#define FLD_S1	m88100opf_S1(opX)
#define FLD_S2	m88100opf_S2(opX)
#define FLD_CR	m88100opf_CR(opX)
#define B5		m88100opf_B5(opX)
#define M5		m88100opf_M5(opX)
#define W5		m88100opf_W5(opX)
#define O5		m88100opf_O5(opX)
#define IMM16	m88100opf_IMM16(opX)
#define VEC9	m88100opf_VEC9(opX)

// 符号拡張・シフト済みのディスプレースメント
// XXX: (負数の右シフト)GCC only
#define D16		(((int32)(int16)opX) << 2)
#define D26		(((int32)(opX << 6) >> 4))
#define IsCO	((opX >> 8) & 1)
#define IsCI	((opX >> 9) & 1)

// レジスタアクセス
#define rD	(r[FLD_D])
#define rS1	(r[FLD_S1])
#define rS2	(r[FLD_S2])
// ダブルワードの2ワード目
#define FLD_D2	((FLD_D + 1) & 0x1f)
#define rD2	(r[FLD_D2])

// FP フィールド
#define FP_T1	m88100opf_FP_T1(opX)
#define FP_T2	m88100opf_FP_T2(opX)
#define FP_TD	m88100opf_FP_TD(opX)

// XXX: 暫定
static float
u2f(uint32 u)
{
	float v;
	memcpy(&v, &u, 4);
	return v;
}

static double
u2d(uint64 u)
{
	double v;
	memcpy(&v, &u, 8);
	return v;
}

static uint32
f2u(float v)
{
	uint32 u;
	memcpy(&u, &v, 4);
	return u;
}

static uint64
d2u(double v)
{
	uint64 u;
	memcpy(&u, &v, 8);
	return u;
}

#define FP_GET(v, t, n)	\
do {	\
	if (t == 0) {	\
		v = u2f(r[n]);	\
	} else if (t == 1) {	\
		if (n == 31) {	\
			FPException(M88K_FPEXCEP_FUNIMP);	\
			return;	\
		}	\
		v = u2d(((uint64)r[n] << 32) | r[n + 1]);	\
	} else {	\
		FPException(M88K_FPEXCEP_FROP);	\
		return;	\
	}	\
} while (0)

#define FP_SET(t, n, v)	\
do {	\
	if (t == 0) {	\
		r[n] = f2u(v);	\
	} else if (t == 1) {	\
		if (n == 31) {	\
			FPException(M88K_FPEXCEP_FUNIMP);	\
			return;	\
		}	\
		uint64 tmp = d2u(v);	\
		r[n] = tmp >> 32;	\
		r[n + 1] = (uint32)tmp;	\
	} else {	\
		FPException(M88K_FPEXCEP_FROP);	\
		return;	\
	}	\
} while (0)

// ブランチ命令入り口
inline void
m88kcpu::EnterBranch(uint32 toaddr)
{
	// ブランチヒストリに登録しつつ、
	// ブランチヒストリからブランチ状況を取得
	auto e = brhist.AddEntry(xip | (IsSuper() ? 1 : 0), toaddr, opX);
	fip = toaddr;

	// STOP 検出。
	// m88k には STOP 状態 (何もせず割り込みだけ待つ状態。m68k の STOP 命令
	// による STOP 状態に相当するもの) は存在せず、割り込みが起きるまで NOP
	// 相当の命令を無限ループで実行し続ける方法が一般的にとられる。
	// 実機であれば (電力が無駄なこと以外は) これで特段困らないかも知れないが
	// エミュレータ (特に高速モード) ではこのような無限ループはホストの CPU
	// がぶん回る状況になるため避けたい。そのためこの無限ループを検出したい。
	// ただし特定の(または定番の)命令列があるわけではなく、任意の無限ループが
	// 用いられているため、汎用的に自動検出する必要がある。

	// ループの 1 回目のブランチで初期化して、
	// 2 回目のブランチで STOP だったかどうかをチェックする。
	// 3 回目以降の場合は STOP でなかったので、もうチェックする必要はない。

	// m68k の STOP 命令が特権命令なので、それに倣ってここでも特権状態のみを
	// 対象とする。

	// 通常のブランチで来る回数がはるかに多いと考えられるので、
	// 先にカウント数をチェック対象かどうか調べる。
	if (e.count < 3 && IsSuper()) {
		if (e.count == 1) {
			// 新規ブランチ
			// 初期化
			nop_counter = 0;
		} else if (pseudo_stop_enable) {
			// 疑似 STOP 状態有効のときに、
			// e.count == 2 のときだけ STOP の検証をすれば良い。

			// d はブランチ間の命令数
			uint32 d = (xip - toaddr) / sizeof(uint32);
			// 遅延ブランチなら 1 命令余分に実行するはずなので 1 足す
			if (opX & (1 << 26)) {
				d++;
			}
			// ブランチ間の命令がすべて NOP なら STOP ということにする。
			if (nop_counter == d) {
				atomic_reqflag |= CPU_REQ_STOP;
			}

			// OpenBSD 6.6 では
			// L1: tb1 #1 r0, #0xff
			//     ld r13, r21, #0x168
			//     bcnd.n eq0, r13, L1
			//     or r2, r25, #0x5ea0
			// の命令列が使用されている。
			// これについては tb1 で実装してある。
		}
		// STOP ではなかったら、あとは放置していい
	}
}

// 通常(即時)ブランチ出口
inline void
m88kcpu::ExitBranch()
{
	// re-fetch branched instruction
	fetch();
}

// 遅延ブランチ出口
// 速度を稼ぐため、ビット反転だけでこれを区別しているので、
// 同じ op 中で複数回呼び出さないこと。
inline void
m88kcpu::ExitDelayBranch()
{
	opF ^= DELAYSLOT;
}

// ロードストアユニットの usr の処理。
// 正常なら true を返す (ので呼び出し側は実行継続)、
// 例外が起きたら処理して false を返すので呼び出し側は即リターンすること。
inline bool
m88kcpu::ldst_usr(uint32& usr)
{
	usr = (opX & (1 << 8)) ? DM_USR : 0;
	if (IsUser() && usr) {
		Exception(M88K_EXCEP_PRIV);
		return false;
	}
	return true;
}

// MEMO: ロードストアでの usr の処理で、CMMU に対して
// usr を引数ではなくSetSuper() の呼び出しで渡すようにしている。
// データのロードストアだけがあるのなら usr を引数で渡すほうが
// 効率的だが、CMMU は 命令/データで同じものを使うため、
// 命令 CMMU では 100% 無駄な引数の引き渡しが必要になってしまう。
// そして命令アクセスのほうが何倍も比率が多い。
// .usr 指定は比率としてはレアなので、状態変更メソッドによる
// 方法にしてみた。

// ロードストアユニットの scale の処理。addr を返す。
inline uint32
m88kcpu::ldst_scale(uint32 size)
{
	if (opX & (1 << 9)) {
		return rS1 + rS2 * size;
	} else {
		return rS1 + rS2;
	}
}

// ロードストアユニットの align の処理。
// 正常なら true を返す (ので呼び出し側は実行継続)、
// 例外が起きたら処理して false を返すので呼び出し側は即リターンすること。
inline bool
m88kcpu::ldst_align(uint32 size, uint32& addr)
{
	if (addr & (size - 1)) {
		if (IsMXM()) {
			addr &= ~(size - 1);
		} else {
			Exception(M88K_EXCEP_MISALIGNED);
			return false;
		}
	}
	return true;
}

// ロードストアユニットの usr と scale の処理 (呼び出し用マクロ)
#define LDST_USR_SCALE(size) do {				\
	if (ldst_usr(usr) == false)					\
		return;									\
	addr = ldst_scale(size);					\
} while (0)

// ロードストアユニットの align の処理 (呼び出し用マクロ)
#define LDST_ALIGN(size) do {					\
	if (ldst_align(size, addr) == false)		\
		return;									\
} while (0)

// ロードストアユニットのまとめたもの
#define LDST_USR_SCALE_ALIGN(size) do {			\
	LDST_USR_SCALE(size);						\
	LDST_ALIGN(size);							\
} while (0)

// 幅規制付き算術左シフト
static inline uint32
ASL(uint32 a, int n)
{
	if (n <= 0)
		return a;
	if (n >= 32)
		return 0;
	return a << n;
}

// 幅規制付き算術右シフト
static inline uint32
ASR(uint32 a, int n)
{
	if (n <= 0)
		return a;
	if (n >= 32)
		return (int32)a < 0 ? -1 : 0;
	// XXX: 負数の右シフト
	return (int32)a >> n;
}

// 幅規制付き論理右シフト
static inline uint32
LSR(uint32 a, int n)
{
	if (n <= 0)
		return a;
	if (n >= 32)
		return 0;
	return a >> n;
}


// 右ローテート
static inline uint32
ROR(uint32 a, int n)
{
	// TODO: 各種コンパイラごとにビルトインがあったりなかったりするらしい
	n &= 31;
	if (n == 0) {
		return a;
	} else {
		return (a >> n) | (a << (32 - n));
	}
}

// add でのオーバーフロー条件
static inline bool
isovf_add(uint32 s1, uint32 s2, uint32 res)
{
	return (int32)((s1 ^ res) & (s2 ^ res)) < 0;
}
// sub でのオーバーフロー条件
static inline bool
isovf_sub(uint32 s1, uint32 s2, uint32 res)
{
	return (int32)((s1 ^ s2) & (s1 ^ res)) < 0;
}

// bcnd のやつ
static inline int
acc_cnd(uint32 r)
{
	int rv;
	if (r & 0x80000000) {
		if (r == 0x80000000) {
			rv = 0x8;
		} else {
			rv = 0x4;
		}
	} else {
		if (r == 0) {
			rv = 0x2;
		} else {
			rv = 0x1;
		}
	}
	return rv;
}

// bitfield
static inline uint32
acc_bf(int w, int ofs)
{
	uint32 m;
	if (w == 0) {
		m = 0xFFFFFFFFU;
	} else {
		m = ~(0xFFFFFFFFU << w);
	}
	return m << ofs;
}

// ここから命令定義

// 000000_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(xmem_bu_imm)
{
	uint32 addr;
	uint64 tmp;

	addr = rS1 + IMM16;
	LDST_ALIGN(1);

	AddCycle(4);	// Table.7-3

	lastaddr = addr;
	tmp = cmmu[1].xmem_8(addr, rD & 0xff);
	if ((int64)tmp < 0) {
		XmemDataException(addr, DM_BU | DM_LOCK);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = tmp;
}

// 000001_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(xmem_w_imm)
{
	uint32 addr;
	uint64 tmp;

	addr = rS1 + IMM16;
	LDST_ALIGN(4);

	AddCycle(4);	// Table.7-3

	lastaddr = addr;
	tmp = cmmu[1].xmem_32(addr, rD);
	if ((int64)tmp < 0) {
		XmemDataException(addr, DM_W | DM_LOCK);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = tmp;
}

// 000010_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(ld_hu_imm)
{
	uint32 addr;
	uint64 data;

	addr = rS1 + IMM16;
	LDST_ALIGN(2);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	data = cmmu[1].load_16(addr);
	if ((int64)data < 0) {
		ReadDataException32(addr, DM_HU);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = data;
}

// 000011_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(ld_bu_imm)
{
	uint32 addr;
	uint64 data;

	addr = rS1 + IMM16;
	LDST_ALIGN(1);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	data = cmmu[1].load_8(addr);
	if ((int64)data < 0) {
		ReadDataException32(addr, DM_BU);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = data;
}

// 000100_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(ld_d_imm)
{
	uint32 addr;
	uint64 data1, data2;

	addr = rS1 + IMM16;
	LDST_ALIGN(8);

	AddCycle(4);	// Table.7-3

	lastaddr = addr;
	data1 = cmmu[1].load_32(addr);
	if ((int64)data1 < 0) {
		ReadDataException64(addr, DM_D1);
		return;
	}
	addr += 4;
	data2 = cmmu[1].load_32(addr);
	if ((int64)data2 < 0) {
		ReadDataException64(addr, DM_D2);
		return;
	}

	rD  = data1;
	rD2 = data2;
	r[0] = 0;
}

// 000101_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(ld_w_imm)
{
	uint32 addr;
	uint64 data;

	addr = rS1 + IMM16;
	LDST_ALIGN(4);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	data = cmmu[1].load_32(addr);
	if ((int64)data < 0) {
		ReadDataException32(addr, DM_W);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = data;
}

// 000110_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(ld_h_imm)
{
	uint32 addr;
	uint64 data;

	addr = rS1 + IMM16;
	LDST_ALIGN(2);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	data = cmmu[1].load_16(addr);
	if ((int64)data < 0) {
		ReadDataException32(addr, DM_H);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = (uint32)(int32)(int16)data;
}

// 000111_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(ld_b_imm)
{
	uint32 addr;
	uint64 data;

	addr = rS1 + IMM16;
	LDST_ALIGN(1);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	data = cmmu[1].load_8(addr);
	if ((int64)data < 0) {
		ReadDataException32(addr, DM_B);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = (uint32)(int32)(int8)data;
}

// 001000_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(st_d_imm)
{
	uint32 addr;
	uint64 rv;

	addr = rS1 + IMM16;
	LDST_ALIGN(8);

	AddCycle(1 + 4);	// Table.7-3

	lastaddr = addr;
	rv = cmmu[1].store_32(addr, rD);
	if ((int64)rv < 0) {
		WriteDataException64(addr, DM_D1);
		return;
	}
	addr += 4;
	rv = cmmu[1].store_32(addr, rD2);
	if ((int64)rv < 0) {
		WriteDataException64(addr, DM_D2);
		return;
	}
}

// 001001_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(st_w_imm)
{
	uint32 addr;
	uint64 rv;

	addr = rS1 + IMM16;
	LDST_ALIGN(4);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	rv = cmmu[1].store_32(addr, rD);
	if ((int64)rv < 0) {
		WriteDataException32(addr, DM_W);
		return;
	}
}

// 001010_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(st_h_imm)
{
	uint32 addr;
	uint64 rv;

	addr = rS1 + IMM16;
	LDST_ALIGN(2);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	rv = cmmu[1].store_16(addr, rD & 0xffff);
	if ((int64)rv < 0) {
		WriteDataException32(addr, DM_H);
		return;
	}
}

// 001011_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(st_b_imm)
{
	uint32 addr;
	uint64 rv;

	addr = rS1 + IMM16;
	LDST_ALIGN(1);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;
	rv = cmmu[1].store_8(addr, rD & 0xff);
	if ((int64)rv < 0) {
		WriteDataException32(addr, DM_B);
		return;
	}
}

// 001100_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(lda_d_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 + IMM16;
}

// 001101_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(lda_w_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 + IMM16;
}

// 001110_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(lda_h_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 + IMM16;
}

// 001111_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(lda_b_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 + IMM16;
}

// 010000_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(and_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & (IMM16 | 0xffff0000U);
}

// 010001_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(and_u_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & ((IMM16 << 16) | 0x0000ffffU);
}

// 010010_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(mask_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & IMM16;
}

// 010011_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(mask_u_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & (IMM16 << 16);
}

// 010100_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(xor_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 ^ IMM16;
}

// 010101_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(xor_u_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 ^ (IMM16 << 16);
}

// 010110_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(or_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 | IMM16;
}

// 010111_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(or_u_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 | (IMM16 << 16);
}

// 011000_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(addu_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 + IMM16;
}

// 011001_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(subu_imm)
{
	if (FLD_D == 0)
		return;
	rD = rS1 - IMM16;
}

// 011010_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(divu_imm)
{
	if (!IsFPUEnable()) {
		fpu_unimpl();
		return;
	}

	AddCycle(38);	// Table.7-6 (ただしざっくり)

	if (IMM16 == 0) {
		Exception(M88K_EXCEP_INT_DIV);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = rS1 / IMM16;
}

// 011011_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(mul_imm)
{
	if (!IsFPUEnable()) {
		fpu_unimpl();
		return;
	}

	AddCycle(4);	// Table.7-6 (ただしざっくり)

	if (FLD_D == 0)
		return;
	rD = rS1 * IMM16;
}

// 011100_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(add_imm)
{
	uint32 imm = IMM16;
	uint32 res = rS1 + imm;
	if (isovf_add(rS1, imm, res)) {
		Exception(M88K_EXCEP_INT_OVF);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = res;
}

// 011101_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(sub_imm)
{
	uint32 imm = IMM16;
	uint32 res = rS1 - imm;
	if (isovf_sub(rS1, imm, res)) {
		Exception(M88K_EXCEP_INT_OVF);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = res;
}

// 011110_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(div_imm)
{
	if (!IsFPUEnable()) {
		fpu_unimpl();
		return;
	}

	AddCycle(38);	// Table.7-6 (ただしざっくり)

	if ((int32)rS1 < 0 || IMM16 == 0) {
		Exception(M88K_EXCEP_INT_DIV);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = rS1 / IMM16;
}

// 011111_DDDDDSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(cmp_imm)
{
	if (FLD_D == 0)
		return;
	rD = acc_cmp(rS1, IMM16);
}

// 100000_DDDDDzzzzz_01000n_nnnnnzzzzz
OP_DEF(ldcr)
{
	if (IsUser()) {
		Exception(M88K_EXCEP_PRIV);
		return;
	}
	if (FLD_CR > countof(cr)) {
		// 88100 undocumented,
		// 88110 compatible
		Exception(M88K_EXCEP_UNIMPL_OP);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = cr[FLD_CR];
}

// 100000_DDDDDzzzzz_01001n_nnnnnzzzzz
OP_DEF(fldcr)
{
	int n = FLD_CR;

	if (IsUser()) {
		// fcr62, 63 はユーザモードでアクセスできる
		// それ以外はスーパバイザ用。
		if (n < 62) {
			Exception(M88K_EXCEP_PRIV);
			return;
		}
	}

	if (8 < n && n < 62) {
		// 未実装 FCR
		// 88100 undocumented,
		// 88110 compatible
		Exception(M88K_EXCEP_UNIMPL_OP);
		return;
	}

	// fcr は 62,63 の配置に細工がしてある
	if (n >= 62) {
		n = n - 62 + 9;
	}
	if (FLD_D == 0)
		return;
	rD = fcr[n];
}

// 100000_zzzzzSSSSS_10000n_nnnnnsssss
OP_DEF(stcr)
{
	if (IsUser()) {
		Exception(M88K_EXCEP_PRIV);
		return;
	}
	if (FLD_CR > countof(cr)) {
		// 88100 undocumented,
		// 88110 unimpl_op
		Exception(M88K_EXCEP_UNIMPL_OP);
		return;
	}
	cr[FLD_CR] = rS1;
	if (FLD_CR == 1) {
		SetPSR();
	}
}

// 100000_zzzzzSSSSS_10001n_nnnnnsssss
OP_DEF(fstcr)
{
	int n = FLD_CR;

	if (IsUser()) {
		// fcr62, 63 はユーザモードでアクセスできる
		// それ以外はスーパバイザ用。
		if (n < 62) {
			Exception(M88K_EXCEP_PRIV);
			return;
		}
	}

	if (8 < n && n < 62) {
		// 未実装 FCR
		// 88100 null op
		// 88110 fpu-unimpl
		return;
	}

	// fcr は 62,63 の配置に細工がしてある
	if (n >= 62) {
		n = n - 62 + 9;
	}
	fcr[n] = rS1;
}

// 100000_DDDDDSSSSS_11000n_nnnnnsssss
OP_DEF(xcr)
{
	uint32 tmp;

	if (IsUser()) {
		Exception(M88K_EXCEP_PRIV);
		return;
	}
	if (FLD_CR > countof(cr)) {
		// 88100 undocumented,
		// 88110 unimpl_op
		Exception(M88K_EXCEP_UNIMPL_OP);
		return;
	}
	tmp = rS1;
	rD = cr[FLD_CR];
	cr[FLD_CR] = tmp;
	if (FLD_CR == 1) {
		SetPSR();
	}
	r[0] = 0;
}

// 100000_DDDDDSSSSS_11001n_nnnnnsssss
OP_DEF(fxcr)
{
	uint32 tmp;
	int n = FLD_CR;

	if (IsUser()) {
		// fcr62, 63 はユーザモードでアクセスできる
		// それ以外はスーパバイザ用。
		if (n < 62) {
			Exception(M88K_EXCEP_PRIV);
			return;
		}
	}

	if (8 < n && n < 62) {
		// 未実装 FCR
		// 88100 null op
		// 88110 fpu-unimpl
		return;
	}

	// fcr は 62,63 の配置に細工がしてある
	if (n >= 62) {
		n = n - 62 + 9;
	}
	tmp = rS1;
	rD = fcr[n];
	fcr[n] = tmp;
	r[0] = 0;
}

// 100001_DDDDDSSSSS_00000n_nnnnnsssss
OP_DEF(fmul)
{
	// 暫定
	double s1;
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s1, FP_T1, FLD_S1);
	FP_GET(s2, FP_T2, FLD_S2);
	FP_SET(FP_TD, FLD_D, s1 * s2);
}

// 100001_DDDDD00000_001000_000nnsssss
OP_DEF(flt)
{
	// 暫定
	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_SET(FP_TD, FLD_D, (double)rS2);
}


// 100001_DDDDDSSSSS_00101n_nnnnnsssss
OP_DEF(fadd)
{
	// 暫定
	double s1;
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s1, FP_T1, FLD_S1);
	FP_GET(s2, FP_T2, FLD_S2);
	FP_SET(FP_TD, FLD_D, s1 + s2);
}

// 100001_DDDDDSSSSS_00110n_nnnnnsssss
OP_DEF(fsub)
{
	// 暫定
	double s1;
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s1, FP_T1, FLD_S1);
	FP_GET(s2, FP_T2, FLD_S2);
	FP_SET(FP_TD, FLD_D, s1 - s2);
}

// 100001_DDDDDSSSSS_00111n_nnnnnsssss
OP_DEF(fcmp)
{
	// 暫定
	double s1;
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s1, FP_T1, FLD_S1);
	FP_GET(s2, FP_T2, FLD_S2);
	rD = acc_fcmp(s1, s2);
}

// 100001_DDDDD00000_010010_0nn00sssss
OP_DEF(int)
{
	// 暫定
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s2, FP_T2, FLD_S2);

	int mode = (fpcr >> 14) & 3;
	switch (mode) {
	 case 0:
		std::fesetround(FE_TONEAREST);
		rD = (uint32)(int32)std::nearbyint(s2);
		break;
	 case 1:
		rD = (uint32)(int32)std::trunc(s2);
		break;
	 case 2:
		rD = (uint32)(int32)std::floor(s2);
		break;
	 case 3:
		rD = (uint32)(int32)std::ceil(s2);
		break;
	}
}

// 100001_DDDDD00000_010100_0nn00sssss
OP_DEF(nint)
{
	// 暫定
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s2, FP_T2, FLD_S2);
	std::fesetround(FE_TONEAREST);
	rD = (uint32)(int32)std::nearbyint(s2);
}

// 100001_DDDDD00000_010110_0nn00sssss
OP_DEF(trnc)
{
	// 暫定
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s2, FP_T2, FLD_S2);
	rD = (uint32)(int32)std::trunc(s2);
}

// 100001_DDDDDSSSSS_01110n_nnnnnsssss
OP_DEF(fdiv)
{
	// 暫定
	double s1;
	double s2;

	if (FLD_D == 0) {
		FPException(M88K_FPEXCEP_FUNIMP);
		return;
	}

	FP_GET(s1, FP_T1, FLD_S1);
	FP_GET(s2, FP_T2, FLD_S2);
	FP_SET(FP_TD, FLD_D, s1 / s2);
}

// 110000_nnnnnnnnnn_nnnnnn_nnnnnnnnnn
OP_DEF(br)
{
	AddCycle(1);	// Table.7-5
	EnterBranch(xip + D26);
	ExitBranch();
}

// 110001_nnnnnnnnnn_nnnnnn_nnnnnnnnnn
OP_DEF(br_n)
{
	AddCycle(0);	// Table.7-5
	EnterBranch(xip + D26);
	ExitDelayBranch();
}

// 110010_nnnnnnnnnn_nnnnnn_nnnnnnnnnn
OP_DEF(bsr)
{
	AddCycle(1);	// Table.7-5
	EnterBranch(xip + D26);
	r[1] = nip;
	ExitBranch();
}

// 110011_nnnnnnnnnn_nnnnnn_nnnnnnnnnn
OP_DEF(bsr_n)
{
	AddCycle(0);	// Table.7-5
	EnterBranch(xip + D26);
	r[1] = nip + 4;
	ExitDelayBranch();
}

// 110100_nnnnnSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(bb0)
{
	if ((rS1 & (1 << B5)) == 0) {
		AddCycle(1);	// Table.7-5
		EnterBranch(xip + D16);
		ExitBranch();
	}
}

// 110101_nnnnnSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(bb0_n)
{
	if ((rS1 & (1 << B5)) == 0) {
		EnterBranch(xip + D16);
		ExitDelayBranch();
	}
}

// 110110_nnnnnSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(bb1)
{
	if ((rS1 & (1 << B5)) != 0) {
		AddCycle(1);	// Table.7-5
		EnterBranch(xip + D16);
		ExitBranch();
	}
}

// 110111_nnnnnSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(bb1_n)
{
	if ((rS1 & (1 << B5)) != 0) {
		EnterBranch(xip + D16);
		ExitDelayBranch();
	}
}

// 111010_MMMMMSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(bcnd)
{
	if (acc_cnd(rS1) & M5) {
		AddCycle(1);	// Table.7-5
		EnterBranch(xip + D16);
		ExitBranch();
	}
}

// 111011_MMMMMSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(bcnd_n)
{
	if (acc_cnd(rS1) & M5) {
		EnterBranch(xip + D16);
		ExitDelayBranch();
	}
}

// 111100_DDDDDSSSSS_100000_wwwwwooooo
OP_DEF(clr_1)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & ~acc_bf(W5, O5);
}

// 111100_DDDDDSSSSS_100010_wwwwwooooo
OP_DEF(set_1)
{
	if (FLD_D == 0)
		return;
	rD = rS1 | acc_bf(W5, O5);
}

// 111100_DDDDDSSSSS_100100_wwwwwooooo
OP_DEF(ext_1)
{
	int w, o, left;
	uint32 tmp;

	if (FLD_D == 0)
		return;
	w = W5;
	if (w == 0)
		w = 32;
	o = O5;
	left = 32 - w - o;
	tmp = ASL(rS1, left);
	tmp = ASR(tmp, left);
	rD = ASR(tmp, o);
}

// 111100_DDDDDSSSSS_100110_wwwwwooooo
OP_DEF(extu_1)
{
	int w, o, left;
	uint32 tmp;

	if (FLD_D == 0)
		return;
	w = W5;
	if (w == 0)
		w = 32;
	o = O5;
	left = 32 - w - o;
	tmp = ASL(rS1, left);
	tmp = LSR(tmp, left);
	rD = tmp >> o;
}

// 111100_DDDDDSSSSS_101000_wwwwwooooo
OP_DEF(mak_1)
{
	if (FLD_D == 0)
		return;
	rD = (rS1 << O5) & acc_bf(W5, O5);
}

// 111100_DDDDDSSSSS_101010_zzzzzooooo
OP_DEF(rot_1)
{
	if (FLD_D == 0)
		return;
	rD = ROR(rS1, O5);
}

// 111100_bbbbbSSSSS_110100_0vvvvvvvvv
OP_DEF(tb0)
{
	if (IsUser() && VEC9 < 128) {
		Exception(M88K_EXCEP_PRIV);
		return;
	}
	if ((rS1 & (1 << B5)) == 0) {
		ExceptionCore(VEC9, ExceptionKind::TRAP);
	}
}

// 111100_bbbbbSSSSS_110110_0vvvvvvvvv
OP_DEF(tb1)
{
	if (IsUser() && VEC9 < 128) {
		Exception(M88K_EXCEP_PRIV);
		return;
	}
	if ((rS1 & (1 << B5)) != 0) {
		ExceptionCore(VEC9, ExceptionKind::TRAP);
	} else if (pseudo_stop_enable) {

		// OpenBSD の STOP エミュレーション。
		// OpenBSD 6.6 では
		// L1: tb1 #1 r0, #0xff
		//     ld r13, r21, #0x168
		//     bcnd.n eq0, r13, L1
		//     or r2, r25, #0x5ea0
		// の命令列が使用されている。
		// tb1 は volatile 変数をアクセスするための同期命令として
		// gcc により出力されている。スーパバイザが volatile を
		// 観測するために同じアドレスの tb1 に直接ブランチで 3 回飛び込んで
		// きたら、それは割り込みか別プロセッサか I/O による変更を待っている
		// はずなので STOP と判断してみる。
		// o. 3回、という数字に確たる理由はない。
		// o. 念の為、16バイト(4命令)後方以内からのブランチに限定しておく。

		if (IsSuper() && VEC9 == 0xff) {
			const auto& e = brhist.entry[brhist.top];
			if (e.count == 3
			 && e.to == xip
			 && e.from - e.to < 16) {
				atomic_reqflag |= CPU_REQ_STOP;
			}
		}
	}
}

// 111100_MMMMMSSSSS_111010_0vvvvvvvvv
OP_DEF(tcnd)
{
	if (IsUser() && VEC9 < 128) {
		Exception(M88K_EXCEP_PRIV);
		return;
	}
	if (acc_cnd(rS1) & M5) {
		ExceptionCore(VEC9, ExceptionKind::TRAP);
	}
}

// 111101_DDDDDSSSSS_000000_xU000sssss
OP_DEF(xmem_bu)
{
	uint64 tmp;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(1);

	AddCycle(4);	// Table.7-3

	lastaddr = addr;

	// 例外安全のために都度 S/U ビットを制御
	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	tmp = cmmu[1].xmem_8(addr, rD & 0xff);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);
	if ((int64)tmp < 0) {
		XmemDataException(addr, DM_BU | DM_LOCK | usr);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = tmp;
}

// 111101_DDDDDSSSSS_000001_xU000sssss
OP_DEF(xmem_w)
{
	uint64 tmp;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(4);

	AddCycle(4);	// Table.7-3

	lastaddr = addr;

	// 例外安全のために都度 S/U ビットを制御
	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	tmp = cmmu[1].xmem_32(addr, rD);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);
	if ((int64)tmp < 0) {
		XmemDataException(addr, DM_W | DM_LOCK | usr);
		return;
	}

	if (FLD_D == 0)
		return;
	rD = tmp;
}

// 111101_DDDDDSSSSS_000010_xU000sssss
OP_DEF(ld_hu)
{
	uint64 data;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(2);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	data = cmmu[1].load_16(addr);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)data < 0) {
		ReadDataException32(addr, DM_HU | usr);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = data;
}

// 111101_DDDDDSSSSS_000011_xU000sssss
OP_DEF(ld_bu)
{
	uint64 data;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(1);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	data = cmmu[1].load_8(addr);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)data < 0) {
		ReadDataException32(addr, DM_BU | usr);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = data;
}

// 111101_DDDDDSSSSS_000100_xU000sssss
OP_DEF(ld_d)
{
	uint64 data1, data2;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(8);

	AddCycle(4);	// Table.7-3

	lastaddr = addr;

	// 例外安全のために都度 S/U ビットを制御
	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	data1 = cmmu[1].load_32(addr);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)data1 < 0) {
		ReadDataException64(addr, DM_D1 | usr);
		return;
	}
	addr += 4;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	data2 = cmmu[1].load_32(addr);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)data2 < 0) {
		ReadDataException64(addr, DM_D2 | usr);
		return;
	}
	rD  = data1;
	rD2 = data2;
	r[0] = 0;
}

// 111101_DDDDDSSSSS_000101_xU000sssss
OP_DEF(ld_w)
{
	uint64 data;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(4);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	data = cmmu[1].load_32(addr);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)data < 0) {
		ReadDataException32(addr, DM_W | usr);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = data;
}

// 111101_DDDDDSSSSS_000110_xU000sssss
OP_DEF(ld_h)
{
	uint64 data;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(2);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	data = cmmu[1].load_16(addr);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)data < 0) {
		ReadDataException32(addr, DM_H | usr);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = (uint32)(int32)(int16)data;
}

// 111101_DDDDDSSSSS_000111_xU000sssss
OP_DEF(ld_b)
{
	uint64 data;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(1);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	data = cmmu[1].load_8(addr);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)data < 0) {
		ReadDataException32(addr, DM_B | usr);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = (uint32)(int32)(int8)data;
}

// 111101_DDDDDSSSSS_001000_xU000sssss
OP_DEF(st_d)
{
	uint64 rv;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(8);

	AddCycle(1 + 4);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	rv = cmmu[1].store_32(addr, rD);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)rv < 0) {
		WriteDataException64(addr, DM_D1 | usr);
		return;
	}

	addr += 4;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	rv = cmmu[1].store_32(addr, rD2);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)rv < 0) {
		WriteDataException64(addr, DM_D2 | usr);
		return;
	}
}

// 111101_DDDDDSSSSS_001001_xU000sssss
OP_DEF(st_w)
{
	uint64 rv;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(4);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	rv = cmmu[1].store_32(addr, rD);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)rv < 0) {
		WriteDataException32(addr, DM_W | usr);
		return;
	}
}

// 111101_DDDDDSSSSS_001010_xU000sssss
OP_DEF(st_h)
{
	uint64 rv;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(2);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	rv = cmmu[1].store_16(addr, rD & 0xffff);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)rv < 0) {
		WriteDataException32(addr, DM_H | usr);
		return;
	}
}

// 111101_DDDDDSSSSS_001011_xU000sssss
OP_DEF(st_b)
{
	uint64 rv;
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE_ALIGN(1);

	AddCycle(3);	// Table.7-3

	lastaddr = addr;

	if (__predict_false(usr)) cmmu[1].SetSuper(false);
	rv = cmmu[1].store_8(addr, rD & 0xff);
	if (__predict_false(usr)) cmmu[1].SetSuper(true);

	if ((int64)rv < 0) {
		WriteDataException32(addr, DM_B | usr);
		return;
	}
}

// 111101_DDDDDSSSSS_001100_xU000sssss
OP_DEF(lda_d)
{
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE(8);

	if (FLD_D == 0)
		return;
	rD = addr;
}

// 111101_DDDDDSSSSS_001101_xU000sssss
OP_DEF(lda_w)
{
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE(4);

	if (FLD_D == 0)
		return;
	rD = addr;
}

// 111101_DDDDDSSSSS_001110_xU000sssss
OP_DEF(lda_h)
{
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE(2);

	if (FLD_D == 0)
		return;
	rD = addr;
}

// 111101_DDDDDSSSSS_001111_xU000sssss
OP_DEF(lda_b)
{
	uint32 addr;
	uint32 usr;

	LDST_USR_SCALE(1);

	if (FLD_D == 0)
		return;
	rD = addr;
}

// 111101_DDDDDSSSSS_010000_00000sssss
OP_DEF(and)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & rS2;
}

// 111101_DDDDDSSSSS_010001_00000sssss
OP_DEF(and_c)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & ~rS2;
}

// 111101_DDDDDSSSSS_010100_00000sssss
OP_DEF(xor)
{
	if (FLD_D == 0)
		return;
	rD = rS1 ^ rS2;
}

// 111101_DDDDDSSSSS_010101_00000sssss
OP_DEF(xor_c)
{
	if (FLD_D == 0)
		return;
	rD = rS1 ^ ~rS2;
}

// 111101_DDDDDSSSSS_010110_00000sssss
OP_DEF(or)
{
	if (FLD_D == 0)	{
		nop_counter++;
		return;
	}
	rD = rS1 | rS2;
}

// 111101_DDDDDSSSSS_010111_00000sssss
OP_DEF(or_c)
{
	if (FLD_D == 0)
		return;
	rD = rS1 | ~rS2;
}

// 111101_DDDDDSSSSS_011000_nn000sssss
OP_DEF(addu)
{
	uint64 res = (uint64)rS1 + rS2;
	if (IsCI) {
		res += GetCY();
	}
	if (IsCO) {
		SetCY(res >> 32);
	}
	if (FLD_D == 0)
		return;
	rD = res;
}

// 111101_DDDDDSSSSS_011001_nn000sssss
OP_DEF(subu)
{
	// sub 命令では CY ビットは Z80 とかと逆で、0 ならボロー発生。
	// 88100 のマニュアルには書いてない。88110 のマニュアルにある。

	// 1 の補数をとって足す。
	uint64 res = (uint64)rS1 + (~rS2);
	if (IsCI) {
		res += GetCY();
	} else {
		res += 1;
	}
	if (IsCO) {
		SetCY(res >> 32);
	}
	if (FLD_D == 0)
		return;
	rD = res;
}

// 111101_DDDDDSSSSS_011010_zz000sssss
OP_DEF(divu)
{
	if (!IsFPUEnable()) {
		fpu_unimpl();
		return;
	}

	AddCycle(38);	// Table.7-6 (ただしざっくり)

	if (rS2 == 0) {
		Exception(M88K_EXCEP_INT_DIV);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = rS1 / rS2;
}

// 111101_DDDDDSSSSS_011011_zz000sssss
OP_DEF(mul)
{
	if (!IsFPUEnable()) {
		fpu_unimpl();
		return;
	}

	AddCycle(4);	// Table.7-6 (ただしざっくり)

	if (FLD_D == 0)
		return;
	rD = rS1 * rS2;
}

// 111101_DDDDDSSSSS_011100_nn000sssss
OP_DEF(add)
{
	uint64 res = (uint64)rS1 + rS2;
	if (IsCI) {
		res += GetCY();
	}
	// XXX: オーバーフローが起きたときにキャリーが更新されるかどうか
	//      ドキュメントされてない気がする
	if (IsCO) {
		SetCY(res >> 32);
	}
	if (isovf_add(rS1, rS2, res)) {
		Exception(M88K_EXCEP_INT_OVF);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = res;
}

// 111101_DDDDDSSSSS_011101_nn000sssss
OP_DEF(sub)
{
	// sub 命令では CY ビットは Z80 とかと逆で、0 ならボロー発生。
	// 88100 のマニュアルには書いてない。88110 のマニュアルにある。

	// 1 の補数をとって足す。
	uint64 res = (uint64)rS1 + (~rS2);
	if (IsCI) {
		res += GetCY();
	} else {
		res += 1;
	}
	// XXX: オーバーフローが起きたときにキャリーが更新されるかどうか
	//      ドキュメントされてない気がする
	if (IsCO) {
		SetCY(res >> 32);
	}
	if (isovf_sub(rS1, rS2, res)) {
		Exception(M88K_EXCEP_INT_OVF);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = res;
}

// 111101_DDDDDSSSSS_011110_zz000sssss
OP_DEF(div)
{
	if (!IsFPUEnable()) {
		fpu_unimpl();
		return;
	}

	AddCycle(38);	// Table.7-6 (ただしざっくり)

	if ((int32)rS1 < 0 || (int32)rS2 <= 0) {
		Exception(M88K_EXCEP_INT_DIV);
		return;
	}
	if (FLD_D == 0)
		return;
	rD = rS1 / rS2;
}

// 111101_DDDDDSSSSS_011111_zz000sssss
OP_DEF(cmp)
{
	if (FLD_D == 0)
		return;
	rD = acc_cmp(rS1, rS2);
}

// 111101_DDDDDSSSSS_100000_00000sssss
OP_DEF(clr_2)
{
	if (FLD_D == 0)
		return;
	rD = rS1 & ~acc_bf((rS2 >> 5) & 0x1f, rS2 & 0x1f);
}

// 111101_DDDDDSSSSS_100010_00000sssss
OP_DEF(set_2)
{
	if (FLD_D == 0)
		return;
	rD = rS1 | acc_bf((rS2 >> 5) & 0x1f, rS2 & 0x1f);
}

// 111101_DDDDDSSSSS_100100_00000sssss
OP_DEF(ext_2)
{
	int w, o, left;
	uint32 tmp;

	if (FLD_D == 0)
		return;
	w = (rS2 >> 5) & 0x1f;
	if (w == 0)
		w = 32;
	o = (rS2     ) & 0x1f;
	left = 32 - w - o;
	tmp = ASL(rS1, left);
	tmp = ASR(tmp, left);
	rD = ASR(tmp, o);
}

// 111101_DDDDDSSSSS_100110_00000sssss
OP_DEF(extu_2)
{
	int w, o, left;
	uint32 tmp;

	if (FLD_D == 0)
		return;
	w = (rS2 >> 5) & 0x1f;
	if (w == 0)
		w = 32;
	o = (rS2     ) & 0x1f;
	left = 32 - w - o;
	tmp = ASL(rS1, left);
	tmp = LSR(tmp, left);
	rD = tmp >> o;
}

// 111101_DDDDDSSSSS_101000_00000sssss
OP_DEF(mak_2)
{
	if (FLD_D == 0)
		return;
	int w = (rS2 >> 5) & 0x1f;
	int o = (rS2     ) & 0x1f;
	rD = (rS1 << o) & acc_bf(w, o);
}

// 111101_DDDDDSSSSS_101010_00000sssss
OP_DEF(rot_2)
{
	if (FLD_D == 0)
		return;
	int o = (rS2    ) & 0x1f;
	rD = ROR(rS1, o);
}

// 111101_zzzzzzzzzz_110000_00000sssss
OP_DEF(jmp)
{
	AddCycle(1);	// Table.7-5
	EnterBranch(rS2);
	ExitBranch();
}

// 111101_zzzzzzzzzz_110001_00000sssss
OP_DEF(jmp_n)
{
	AddCycle(0);	// Table.7-5
	EnterBranch(rS2);
	ExitDelayBranch();
}

// 111101_zzzzzzzzzz_110010_00000sssss
OP_DEF(jsr)
{
	AddCycle(1);	// Table.7-5
	EnterBranch(rS2);
	r[1] = nip;
	ExitBranch();
}

// 111101_zzzzzzzzzz_110011_00000sssss
OP_DEF(jsr_n)
{
	AddCycle(0);	// Table.7-5
	EnterBranch(rS2);
	r[1] = nip + 4;
	ExitDelayBranch();
}

// 111101_DDDDDzzzzz_111010_00000sssss
OP_DEF(ff1)
{
	int i;
	uint32 m = 0x80000000U;

	if (FLD_D == 0)
		return;

	if (rS2 == 0) {
		rD = 32;
	} else {
		for (i = 31; i >= 0; i--, m >>= 1) {
			if ((rS2 & m) != 0) {
				break;
			}
		}
		rD = i;
	}

}

// 111101_DDDDDzzzzz_111011_00000sssss
OP_DEF(ff0)
{
	int i;
	uint32 m = 0x80000000U;

	if (FLD_D == 0)
		return;

	if (rS2 == 0xffffffffU) {
		rD = 32;
	} else {
		for (i = 31; i >= 0; i--, m >>= 1) {
			if ((rS2 & m) == 0) {
				break;
			}
		}
		rD = i;
	}
}

// 111101_zzzzzzzzzz_111111_0000000000
OP_DEF(rte)
{
	AddCycle(2);	// Table.7-5

	SetPSR(epsr);

	// LUNA88K の ROM1.2 は sxip, snip, sfip を制御してくる。
	// ROM1.2 の DAE ハンドラは、
	// sxip = INVALID
	// snip = 0 (INVALID)
	// sfip = original sxip
	// で rte してくる。
	// CPU は sxip は評価しない。snip は 0 なので、無効として nop 扱い。
	// (実際はクロックが消費されるかも?)
	// sfip が VALID なので、そこから実行を再開する。
	if (snip & SIP_V) {
		EnterBranch(snip & SIP_MASK);
		ExitBranch();
		fip = sfip & SIP_MASK;
	} else {
		EnterBranch(sfip & SIP_MASK);
		ExitBranch();
	}

	// ディレイスロットにいた命令かどうかをやんわり復元する。
	// 88110 ではこの情報が E?IP に保持されているが
	// 88100 にはない。
	if (nip + 4 != fip) {
		ExitDelayBranch();
	}
}

// 111101_zzzzzSSSSS_111110_00000SSSSS
OP_DEF(tbnd_1)
{
	if (rS1 > rS2) {
		Exception(M88K_EXCEP_BOUNDS);
	}
}

// 111110_zzzzzSSSSS_nnnnnn_nnnnnnnnnn
OP_DEF(tbnd_2)
{
	if (rS1 > IMM16) {
		Exception(M88K_EXCEP_BOUNDS);
	}
}

// 111111_zzzzzzzzzz_000011_zzzzzzzzzz
OP_DEF(getc)
{
	printf("getc inst\n");
	int c = getchar();
	r[2] = c;
}

// 111111_zzzzzzzzzz_000100_zzzzzzzzzz
OP_DEF(putc)
{
	// r3 を表示
	putchar(r[3]);
}
