//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "device.h"
#include "vm.h"

// 全デバイスリスト
std::vector<Device *> gDevices;

//
// Device
//

// コンストラクタ
Device::Device(const std::string& objname_)
	: inherited(objname_)
{
	gDevices.push_back(this);
}

// デストラクタ
Device::~Device()
{
	// 自分自身をリストから削除する
	// XXX マルチスレッドな状況からは呼ばれないはず?
	for (auto it = gDevices.begin(); it != gDevices.end(); ++it) {
		if (*it == this) {
			gDevices.erase(it);
			break;
		}
	}
}

bool
Device::Create()
{
	return true;
}

bool
Device::Init()
{
	return true;
}

bool
Device::Apply()
{
	return true;
}

bool
Device::PowerOn()
{
	return true;
}

bool
Device::PowerOff()
{
	return true;
}

// 本体リセットスイッチによるリセット
void
Device::ResetHard()
{
}


//
// IODevice
//

// コンストラクタ
IODevice::IODevice(const std::string& objname_)
	: inherited(objname_)
{
	devaddr = 0;
	devlen  = 0;
}

// デストラクタ
IODevice::~IODevice()
{
}

uint64
IODevice::Read8(uint32 addr)
{
	return 0xff;
}

uint64
IODevice::Read16(uint32 addr)
{
	return (Read8(addr) << 8) | Read8(addr + 1);
}

uint64
IODevice::Read32(uint32 addr)
{
	return (Read16(addr) << 16) | Read16(addr + 2);
}

uint64
IODevice::Write8(uint32 addr, uint32 data)
{
	return 0;
}

uint64
IODevice::Write16(uint32 addr, uint32 data)
{
	int64 berr;

	berr = Write8(addr, data >> 8);
	if (berr < 0)
		return berr;
	berr = Write8(addr + 1, data & 0xff);
	if (berr < 0)
		return berr;
	return 0;
}

uint64
IODevice::Write32(uint32 addr, uint32 data)
{
	int64 berr;

	berr = Write16(addr, data >> 16);
	if (berr < 0)
		return berr;
	berr = Write16(addr + 2, data & 0xffff);
	if (berr < 0)
		return berr;
	return 0;
}

uint64
IODevice::Peek8(uint32 addr)
{
	return 0xff;
}


//
// バスエラー
//

std::unique_ptr<BusErrDevice> gBusErr;

// コンストラクタ
BusErrDevice::BusErrDevice()
	: inherited("(BusErr)")
{
	ClearAlias();
	AddAlias("BusErr");

	devaddr = -1;	// XXX どうしたもんか
}

BusErrDevice::~BusErrDevice()
{
}

uint64
BusErrDevice::Read8(uint32 addr)
{
	putlog(1, "$%08x.b", addr);
	return (uint64)-1;
}

uint64
BusErrDevice::Read16(uint32 addr)
{
	putlog(1, "$%08x.w", addr);
	return (uint64)-1;
}

uint64
BusErrDevice::Read32(uint32 addr)
{
	putlog(1, "$%08x.l", addr);
	return (uint64)-1;
}

uint64
BusErrDevice::Write8(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.b <- %02X", addr, data);
	return (uint64)-1;
}

uint64
BusErrDevice::Write16(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.w <- %04X", addr, data);
	return (uint64)-1;
}

uint64
BusErrDevice::Write32(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.l <- %08X", addr, data);
	return (uint64)-1;
}

uint64
BusErrDevice::Peek8(uint32 addr)
{
	return (uint64)-1;
}


//
// 何もしないデバイス
//

std::unique_ptr<NopIODevice> gNopIO;

// コンストラクタ
NopIODevice::NopIODevice()
	: inherited("(NopIO)")
{
	ClearAlias();
	AddAlias("NopIO");
}

NopIODevice::~NopIODevice()
{
}

uint64
NopIODevice::Read8(uint32 addr)
{
	putlog(1, "$%08x.b", addr);
	return 0xff;
}

uint64
NopIODevice::Read16(uint32 addr)
{
	putlog(1, "$%08x.w", addr);
	return 0xffff;
}

uint64
NopIODevice::Read32(uint32 addr)
{
	putlog(1, "$%08x.l", addr);
	return 0xffffffff;
}

uint64
NopIODevice::Write8(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.b <- $%02x", addr, data);
	return 0;
}

uint64
NopIODevice::Write16(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.w <- $%04x", addr, data);
	return 0;
}

uint64
NopIODevice::Write32(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.l <- $%08x", addr, data);
	return 0;
}

uint64
NopIODevice::Peek8(uint32 addr)
{
	return 0xff;
}
