//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "device.h"
#include "qvector.h"
#include <array>
#include <mutex>
#include <queue>
#include <vector>

//          IODevice
//              |
//              v
//        EthernetDevice (イーサネットとしての共通部分)
//              |
//     +--------+-------+
//     v                v
// LanceDevice     RTL8019Device
// (LUNA, AM7990)  (X68000, RTL8019)
//
// これ自体は IODevice である必要はないが、LanceDevice などの親になるため
// IODevice にしてある。(多重継承はしない)

class NetDriver;

// MAC アドレス格納用のクラス
class macaddr_t : public std::array<uint8, 6>
{
	using inherited = std::array<uint8, 6>;
 public:
	// この MAC アドレスが 00:00:00:00:00:00 なら true
	bool empty() const {
		for (const auto& v : *this) {
			if (v != 0) {
				return false;
			}
		}
		return true;
	}

	// 区切り文字なしの文字列形式を返す。主に ROM 埋め込み用で英字は大文字。
	// 01:23:AB なら "0123AB" のような感じ。
	std::string to_string() const {
		std::string buf;
		for (const auto& v : *this) {
			buf += string_format("%02X", v);
		}
		return buf;
	}

	// sep を区切り文字とする文字列形式を返す。主に表示用で英字は小文字。
	// 01:23:AB で sep = ":" なら "01:23:ab" のような感じ。
	std::string to_string(char sep) const {
		std::string buf;
		for (const auto& v : *this) {
			buf += string_format("%02x%c", v, sep);
		}
		buf.pop_back();
		return buf;
	}
};

// 受信パケット
using RXPacket = std::unique_ptr<qvector<uint8>>;
// 受信パケットキュー用
using RXPacketQueue = std::queue<RXPacket>;

class EthernetDevice : public IODevice
{
	using inherited = IODevice;
 protected:
	EthernetDevice(const std::string& objname_);
 public:
	virtual ~EthernetDevice() override;

	bool Create() override;

	// パケットを受信した。
	// ホストネットワークドライバから呼ばれる。
	// バッファをコピーしてこちらのスレッドに処理を引き渡すところまでを行う。
	// packet はプリアンブルを含まず、イーサネットフレームヘッダ(14バイト)、
	// データ(46-1500バイト)、FCS(CRC) (4バイト) からなる。
	// 渡されたパケットの所有権はこちらで引き取る。
	bool RecvPacket(RXPacket packet);

	// MAC アドレスを取得
	macaddr_t GetMacAddr() const {
		return macaddr;
	}

	// 統計情報の取得
	uint64 GetTXPkts() const { return tx_pkts; }
	uint64 GetRXPkts() const { return rx_pkts; }

 protected:
	// このデバイスと継承デバイスのメンバ変数を読み書きする場合は
	// 必ずこのロックを取得する。?
	std::mutex mtx {};

	// パケット受信許可。
	bool RecvEnabled {};

	// MAC アドレスを生成する
	static void GenerateMacAddr(macaddr_t *addr);

	// MAC アドレス文字列をバイナリに変換する
	static bool ParseMacAddr(macaddr_t& dst, const std::string& src);

	// パケットを送信する
	bool SendPacket(const std::vector<uint8>& buf);

	// 継承クラスはこのメソッドをオーバーライドして、受信に対する処理を行う。
	// パケットを受信してキューに投入した後に呼び出される。
	virtual void NotifyRecvPacket() = 0;

	// MAC アドレス
	macaddr_t macaddr {};

	// 受信パケットキュー
	// VMとホストの通信速度調停のためキューが必要。
	RXPacketQueue rxq {};

	// 統計情報
	uint64 tx_pkts {};
	uint64 rx_pkts {};

 private:
	// Create() の下請け
	bool CreateNone();
	bool CreateAFPacket();
	bool CreateTap();
	bool CreateBPF();
};

extern std::unique_ptr<EthernetDevice> gEthernet;
