//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// キーボードとマウス
//
// キーボード/マウスからシリアルポートへの入力は、1バイトの送信自体に約1msec
// (9600bps、スタートビット1、ストップビット1) で送信後に 1msec 空けるそう
// なので、ここでは単純に 2msec を1スロットとする。
// マウスは 20msec ごとにサンプリングして連続する3バイトを送り、
// キーボードは任意のスロットで 1バイト送ってくるので、概念的には
// 10スロットのうち 0, 1, 2 をマウスが、3..9 をキーボードが使用する
// ということにする。

#include "lunakbd.h"
#include "sio.h"

// コンストラクタ
LunaKeyboard::LunaKeyboard()
{
	led.resize(2);

	key_event.func = (DeviceCallback_t)&LunaKeyboard::KeyboardCallback;
	key_event.time = 0;
	key_event.SetName("Keyboard Slot");

	// マウスイベント#0 はサンプリング用も兼ねているので常にオン
	mouse_event[0].dev = this;
	mouse_event[0].func =
		(DeviceCallback_t)&LunaKeyboard::MouseSamplingCallback;
	mouse_event[0].time = 20_msec;
	mouse_event[0].SetName("Mouse Sampling");

	for (int i = 1; i < 3; i++) {
		mouse_event[i].dev = this;
		mouse_event[i].func = (DeviceCallback_t)&LunaKeyboard::MouseCallback;
		mouse_event[i].time = 2_msec * i;
		mouse_event[i].SetName(string_format("Mouse Slot%d", i));
	}

	monitor.func = (MonitorCallback_t)&LunaKeyboard::MonitorUpdate;
	monitor.SetSize(30, 1);
	monitor.Regist(ID_MONITOR_KEYBOARD);
}

// デストラクタ
LunaKeyboard::~LunaKeyboard()
{
}

// 電源オン
bool
LunaKeyboard::PowerOn()
{
	// XXX リセットはマウスには届かないような気がするので
	// 電源オンで初期化。
	mouse_x = 0;
	mouse_y = 0;
	mouse_r = false;
	mouse_l = false;
	mouse_m = false;
	prev_r = false;
	prev_l = false;
	prev_m = false;

	return true;
}

// リセット
void
LunaKeyboard::ResetHard()
{
	inherited::ResetHard();

	// 動作中のタイマーは一旦とめる
	key_event.Stop();
	for (int i = 0; i < countof(mouse_event); i++) {
		mouse_event[i].Stop();
	}

	// で、マウスサンプリングは電源オン時から有効らしい
	mouse_event[0].Start();
}

// モニター
void
LunaKeyboard::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int x, y;
	screen.Clear();

	x = mouse_x;
	y = mouse_y;
	if (x < -999)
		x = -999;
	if (x > 999)
		x = 999;
	if (y < -999)
		y = -999;
	if (y > 999)
		y = 999;
	screen.Print(0, 0, "Mouse: x=%-4d y=%-4d %c %c %c",
		x, y,
		(mouse_l ? 'L' : '-'),
		(mouse_m ? 'M' : '-'),
		(mouse_r ? 'R' : '-'));
}

// 共通キーコードを LUNA キーコードに変換して返す。
// 対応する LUNA のキーがなければ 0 を返す。
uint
LunaKeyboard::KeyToMDKey(uint keycode) const
{
	if (keycode >= countof(keycode2lunakey_table)) {
		return 0;
	}

	return keycode2lunakey_table[keycode];
}

// シフト状態なら true を返す
bool
LunaKeyboard::IsShift() const
{
	return pressed[KC_SHIFT_L] || pressed[KC_SHIFT_R];
}

// キー入力を1つ処理する
void
LunaKeyboard::KeyInput(uint keystat)
{
	uint keycode = KC_CODE(keystat);
	bool ismake = KC_IS_MAKE(keystat);

	assert(keycode < KC_max);

	// 共通処理へ
	if (KeyInputCommon(keystat) == false) {
		return;
	}

	// LED キーの処理
	if (ismake) {
		// LED はトグル?
		int n = Keycode2LED(keycode);
		if (n >= 0) {
			led[n] = !led[n];
		}
	}

	// LUNA のキーボードはキーボード側でキーリピートしないので
	// ここでは何もしなくてよい
}

// 共通キーコードから LED 番号を返す。LED キーでなければ -1 を返す。
int
LunaKeyboard::Keycode2LED(uint keycode) const
{
	// 2つしかないのでこのくらいでいいか
	if (keycode == KC_kana)
		return 0;
	if (keycode == KC_CAPS)
		return 1;

	return -1;
}

// マウス入力
void
LunaKeyboard::MouseInput(int x, int y, bool rb, bool lb, bool mb)
{
	// ホストからマウスイベントのたびに呼ばれるので、ここで積算。
	// 20msec ごとのサンプリングのタイミングで積算結果を処理する。
	// LUNA のマウスの Y 方向は上に動く方向がプラスなので、ホスト入力とは逆。
	mouse_x += x;
	mouse_y -= y;
	mouse_r = rb;
	mouse_l = lb;
	mouse_m = mb;
}

// キーイベントを起こせる空きスロットを計算して、イベントを開始
void
LunaKeyboard::SendStart()
{
	// 例えば、スロット 4 期間内のどこかでキー入力(K)が起きると
	// スロット 5 の先頭時刻にイベント(E)を起こす。
	//
	//   通算スロット  44    45    46
	//   スロット番号   4     5     6
	//               +-----+-----+-----+..
	//               | K--->E
	//               +-----+-----+-----+..
	//
	// スロット 9 期間内のどこかでキー入力(K)が起きると、(0, 1, 2 はマウス用
	// のスロットにしているので、これを飛ばして) スロット 3 の先頭でイベント
	// (E)を起こす。
	//
	//   通算スロット  209   210   211   212   213
	//   スロット番号    9     0     1     2     3
	//               +-----+-----+-----+-----+-----+..
	//               |    K------------------>E
	//               +-----+-----+-----+-----+-----+..

	// イベントが進行中ならなにもせず帰る
	if (key_event.IsRunning()) {
		return;
	}

	uint64 now = gMPU->GetVirtTime();

	// t は現在の仮想時刻に対応する通算スロット
	uint64 t = now / 2_msec;
	// n はその次の通算スロット
	uint64 n = t + 1;

	// マウススロットなら避ける
	while (n % 10 < 3) {
		n++;
	}

	// n の時刻がイベントを起こす時刻
	uint64 target = n * 2_msec;

	// イベントを登録。
	key_event.time = target - now;
	key_event.Start();
}

// キー入力コールバック
// SIO に送信するタイミングで呼ばれる。
void
LunaKeyboard::KeyboardCallback(Event& ev)
{
	// 呼ばれたということはキューは空でないはず
	assert(sendqueue.Empty() == false);

	// キューから取り出す。ここで取り出せるのは keystat
	uint keystat = sendqueue.Dequeue();
	uint keycode = KC_CODE(keystat);
	// ここで LUNA キーコードに変換 (変換できないはずはない)
	uint mdkey = KeyToMDKey(keycode);
	assert(mdkey != 0);

	// 送信
	if (KC_IS_BREAK(keystat)) {
		mdkey |= 0x80;
	}
	gSIO->Rx(1, mdkey);

	if (Keycode2LED(keystat) >= 0) {
		// かなと CAP キーの押下なら、続けてキーの復旧も送る。
		// 押下の時だけここに来ればいいので keycode じゃなく keystat。

		// (BreakKey() が内部でもう一度イベントをキックしている)
		BreakKey(keycode);
	}

	// キューにまだあるならここで次のイベントをキックする必要がある
	if (sendqueue.Empty() == false) {
		SendStart();
	}
}

// マウスサンプリングコールバック
// 20msec ごとに呼ばれる。
void
LunaKeyboard::MouseSamplingCallback(Event& ev)
{
	int reset_x = 0;
	int reset_y = 0;

	// 移動量±1は送信しないが次回に向けた積算は続けるそうだ。
	if (mouse_x == -1 || mouse_x == 1) {
		reset_x = mouse_x;
		mouse_x = 0;
	}
	if (mouse_y == -1 || mouse_y == 1) {
		reset_y = mouse_y;
		mouse_y = 0;
	}

	// 移動量が±1 を超えるかボタンがどれかでも変化すれば送信。
	if (mouse_x != 0 ||
	    mouse_y != 0 ||
	    mouse_r != prev_r ||
	    mouse_l != prev_l ||
	    mouse_m != prev_m)
	{
		putlog(2, "Mouse Sampling x=%d y=%d %c%c%c",
			mouse_x, mouse_y,
			(mouse_r ? 'L' : '-'),
			(mouse_m ? 'M' : '-'),
			(mouse_l ? 'R' : '-'));

		// 1バイト目、ボタン状態。
		// mouse_[rlm] 変数は押し下げが true、LUNA のマウスは押し下げが %0
		uint8 b = 0x80 |
			(mouse_l ? 0 : 0x04) |
			(mouse_m ? 0 : 0x02) |
			(mouse_r ? 0 : 0x01);

		// 2バイト目が X、3バイト目が Y。
		// 移動量は絶対値 0x7f でサチる (signed char ではない)。
		if (mouse_x < -127)
			mouse_x = -127;
		if (mouse_x > 127)
			mouse_x = 127;

		if (mouse_y < -127)
			mouse_y = -127;
		if (mouse_y > 127)
			mouse_y = 127;

		// 1バイト目はここで送信
		gSIO->Rx(1, b);

		// 2バイト目は 2msec 秒後に送信
		mouse_event[1].code = (uint8)mouse_x;
		mouse_event[1].Start();

		// 3バイト目は 4msec 秒後に送信
		mouse_event[2].code = (uint8)mouse_y;
		mouse_event[2].Start();
	} else {
		putlog(3, "Mouse Sampling inactive");
	}

	// 積算リセット
	mouse_x = reset_x;
	mouse_y = reset_y;
	prev_r = mouse_r;
	prev_l = mouse_l;
	prev_m = mouse_m;

	// 20msec 後に再び自分を呼び出す
	ev.Start();
}

// マウスの後続バイトを送出するイベントコールバック
void
LunaKeyboard::MouseCallback(Event& ev)
{
	gSIO->Rx(1, ev.code);
}

// ホストからの制御
void
LunaKeyboard::Command(uint32 data)
{
}


// キーコード変換表
// 共通キーコードから LUNA キーコードに変換する。
// ぶっちゃけた話共通キーコードは LUNA キーコードをベースにしているので、
// 穴がある以外は全部透過。
/*static*/ const uint
LunaKeyboard::keycode2lunakey_table[0x80] = {
	NoKey,					// [00]
	NoKey,					// [01]
	NoKey,					// [02]
	NoKey,					// [03]
	NoKey,					// [04]
	NoKey,					// [05]
	NoKey,					// [06]
	NoKey,					// [07]
	NoKey,					// [08]
	LunaKey_TAB,			// [09]
	LunaKey_CTRL,			// [0a]
	LunaKey_kana,			// [0b]
	LunaKey_SHIFT_L,		// [0c]
	LunaKey_SHIFT_R,		// [0d]
	LunaKey_CAPS,			// [0e]
	LunaKey_SF,				// [0f]

	LunaKey_ESC,			// [10]
	LunaKey_BS,				// [11]
	LunaKey_enter,			// [12]
	NoKey,					// [13]
	LunaKey_space,			// [14]
	LunaKey_DEL,			// [15]
	LunaKey_XFER,			// [16]
	LunaKey_VALID,			// [17]
	LunaKey_PF11,			// [18]
	LunaKey_PF12,			// [19]
	LunaKey_PF13,			// [1a]
	LunaKey_PF14,			// [1b]
	LunaKey_up,				// [1c]
	LunaKey_left,			// [1d]
	LunaKey_right,			// [1e]
	LunaKey_down,			// [1f]

	NoKey,					// [20]
	NoKey,					// [21]
	LunaKey_1,				// [22]
	LunaKey_2,				// [23]
	LunaKey_3,				// [24]
	LunaKey_4,				// [25]
	LunaKey_5,				// [26]
	LunaKey_6,				// [27]
	LunaKey_7,				// [28]
	LunaKey_8,				// [29]
	LunaKey_9,				// [2a]
	LunaKey_0,				// [2b]
	LunaKey_minus,			// [2c]
	LunaKey_circum,			// [2d]
	LunaKey_backslash,		// [2e]
	NoKey,					// [2f]

	NoKey,					// [30]
	NoKey,					// [31]
	LunaKey_Q,				// [32]
	LunaKey_W,				// [33]
	LunaKey_E,				// [34]
	LunaKey_R,				// [35]
	LunaKey_T,				// [36]
	LunaKey_Y,				// [37]
	LunaKey_U,				// [38]
	LunaKey_I,				// [39]
	LunaKey_O,				// [3a]
	LunaKey_P,				// [3b]
	LunaKey_at,				// [3c]
	LunaKey_bracketleft,	// [3d]
	NoKey,					// [3e]
	NoKey,					// [3f]

	NoKey,					// [40]
	NoKey,					// [41]
	LunaKey_A,				// [42]
	LunaKey_S,				// [43]
	LunaKey_D,				// [44]
	LunaKey_F,				// [45]
	LunaKey_G,				// [46]
	LunaKey_H,				// [47]
	LunaKey_J,				// [48]
	LunaKey_K,				// [49]
	LunaKey_L,				// [4a]
	LunaKey_semicolon,		// [4b]
	LunaKey_colon,			// [4c]
	LunaKey_bracketright,	// [4d]
	NoKey,					// [4e]
	NoKey,					// [4f]

	NoKey,					// [50]
	NoKey,					// [51]
	LunaKey_Z,				// [52]
	LunaKey_X,				// [53]
	LunaKey_C,				// [54]
	LunaKey_V,				// [55]
	LunaKey_B,				// [56]
	LunaKey_N,				// [57]
	LunaKey_M,				// [58]
	LunaKey_comma,			// [59]
	LunaKey_period,			// [5a]
	LunaKey_slash,			// [5b]
	LunaKey_underscore,		// [5c]
	NoKey,					// [5d]
	NoKey,					// [5e]
	NoKey,					// [5f]

	NoKey,					// [60]
	LunaKey_PAD_plus,		// [61]
	LunaKey_PAD_minus,		// [62]
	LunaKey_PAD_7,			// [63]
	LunaKey_PAD_8,			// [64]
	LunaKey_PAD_9,			// [65]
	LunaKey_PAD_4,			// [66]
	LunaKey_PAD_5,			// [67]
	LunaKey_PAD_6,			// [68]
	LunaKey_PAD_1,			// [69]
	LunaKey_PAD_2,			// [6a]
	LunaKey_PAD_3,			// [6b]
	LunaKey_PAD_0,			// [6c]
	LunaKey_PAD_decimal,	// [6d]
	LunaKey_PAD_enter,		// [6e]
	NoKey,					// [6f]

	NoKey,					// [70]
	NoKey,					// [71]
	LunaKey_F1,				// [72]
	LunaKey_F2,				// [73]
	LunaKey_F3,				// [74]
	LunaKey_F4,				// [75]
	LunaKey_F5,				// [76]
	LunaKey_F6,				// [77]
	LunaKey_F7,				// [78]
	LunaKey_F8,				// [79]
	LunaKey_F9,				// [7a]
	LunaKey_F10,			// [7b]
	LunaKey_PAD_multiply,	// [7c]
	LunaKey_PAD_divide,		// [7d]
	LunaKey_PAD_equal,		// [7e]
	LunaKey_PAD_comma,		// [7f]
};
