//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// MPU (m680x0)

#pragma once

#include "event.h"
#include "mpu.h"
#include "m68030.h"

class MPU680x0Device : public MPUDevice
{
	using inherited = MPUDevice;
 public:
	MPU680x0Device();
	virtual ~MPU680x0Device() override;

	bool Init() override;
	bool PowerOn() override;
	void ResetHard() override;

	// MPU を vtime [nsec] 分実行する
	uint32 Run(uint32 vtime) override;

	// 現在の仮想時刻 [nsec] を取得する
	uint64 GetVirtTime() const override;

	// 現在の PPC を取得
	uint32 GetPPC() const override { return RegPPC; }

	// 現在アクセス中の論理アドレス、物理アドレスを取得。
	uint32 GetLaddr() const override { return cpu->bus.laddr; }
	uint32 GetPaddr() const override { return cpu->bus.paddr; }

	// アクセスウェイトを加算
	void AddCycle(int32 cycle) override { cpu->used_cycle += cycle; }

	// 割り込みレベルが変わったことを MPU に通知
	void Interrupt(int level) override;

	// MPU の処理をこの命令で中断させる。
	// スケジューラのイベントリストに追加が発生した時に VM 側から
	// MPU へ処理の中断を指示する。
	void Release() override { cpu->atomic_reqflag |= CPU_REQ_RELEASE; }

	// A-Line, F-Line 命令エミュレーションのコールバックを指定する。
	// (Human68k エミュレーションで使用)
	void SetALineCallback(bool (*callback)(m68kcpu*, void*), void *arg);
	void SetFLineCallback(bool (*callback)(m68kcpu*, void*), void *arg);

	// ダブルバスフォールトのコールバックを指定する。
	void SetHaltCallback(void (*callback)(void*), void *arg);

	// FPU を持っているか
	bool HaveFPU() const {
		return cpu->has_fpu;
	}

	// CPU コアを取得 (主にデバッグ用途)
	m68kcpu *GetCPU() const {
		return cpu;
	}

 private:
	// イベントコールバック
	void Callback(Event& ev);

	// 割り込み用イベント
	Event event { this };

	// レジスタモニター
	DECLARE_MONITOR_CALLBACK(MonitorUpdateReg);
	Monitor reg_monitor { this };

	// ATC モニター
	DECLARE_MONITOR_CALLBACK(MonitorUpdateATC);
	Monitor atc_monitor { this };

	// 割り込み確定のためのディレイ [nsec]
	uint64 intdelay {};

	m68kcpu *cpu {};
};

#define gMPU680x0	(dynamic_cast<MPU680x0Device *>(gMPU.get()))
