//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "device.h"
#include "scheduler.h"

// Intel 82C55 の仕様書にはコントロールポートからの読み出しは明記してある。
// NEC uPD8255 の仕様書にはコントロールポートからの読み出しについての記載が
// ない (禁止とも書いてない)。
class i8255Device : public IODevice
{
	using inherited = IODevice;

 public:
	i8255Device(const std::string& objname_);
	virtual ~i8255Device() override;

	void ResetHard() override;

 protected:
	// コントロールポートの読み出し
	uint32 ReadCtrl() const { return ctrl; }
	// コントロールポートへの書き込み
	void WriteCtrl(uint32);
	// PortC への個別書き込み
	virtual void SetPC(int, int);

	// 7  6  5  4  3   2  1  0
	// 1  AM AM AD CHD BM BD CLD
	//    |  |  |  |   |  |  +--- ポートC下位グループの方向。入力なら1
	//    |  |  |  |   |  +------ ポートB の方向。入力なら1
	//    |  |  |  |   +--------- ポートB のモード。
	//    |  |  |  +------------- ポートC上位グループの方向。入力なら1
	//    |  |  +---------------- ポートA の方向。入力なら1
	//    +--+------------------- ポートA のモード。
	uint8 ctrl = 0;			// コントロールポートの値

	static const uint DIR_IN  = 1;
	static const uint DIR_OUT = 0;
	// グループA/B (ポートA/B) のモード
	uint GetGroupAMode() const	{ return (ctrl >> 5) & 3; }
	uint GetGroupBMode() const	{ return (ctrl >> 2) & 1; }
	// ポート A/B/C上位/C下位 の入出力方向
	uint PortADir() const	{ return (ctrl & 0x10) ? DIR_IN : DIR_OUT; }
	uint PortBDir() const	{ return (ctrl & 0x02) ? DIR_IN : DIR_OUT; }
	uint PortCHDir() const	{ return (ctrl & 0x08) ? DIR_IN : DIR_OUT; }
	uint PortCLDir() const	{ return (ctrl & 0x01) ? DIR_IN : DIR_OUT; }
};


class PPIDevice : public i8255Device
{
	using inherited = i8255Device;
 private:
	static const int baseaddr = 0xe9a000;

 public:
	PPIDevice();
	virtual ~PPIDevice() override;

 protected:
	// BusIO インタフェース
	static const uint32 NPORT = 4;
	uint64 Read(uint32 offset);
	uint64 Write(uint32 offset, uint32 data);
	uint64 Peek(uint32 offset);
};

class PIO0Device : public i8255Device
{
	using inherited = i8255Device;
 public:
	PIO0Device();
	virtual ~PIO0Device() override;

	bool Init() override;

	// DIP-SW1 の状態を取得 (ROM エミュレーション用)
	//
	//    b7     b6     b5     b4     b3     b2     b1     b0
	// +------+------+------+------+------+------+------+------+
	// | #1-8 | #1-7 | #1-6 | #1-5 | #1-4 | #1-3 | #1-2 | #1-1 |
	// +------+------+------+------+------+------+------+------+
	// %1 が UP、%0 が DOWN
	uint32 GetDIPSW1() const { return GetPA(); }

 protected:
	// BusIO インタフェース
	static const uint32 NPORT = 4;
	uint64 Read(uint32 offset);
	uint64 Write(uint32 offset, uint32 data);
	uint64 Peek(uint32 offset);

 private:
	uint32 GetPA() const;
	uint32 GetPB() const;
	uint32 GetPC() const;
	void SetPC(int, int) override;

	DECLARE_MONITOR_CALLBACK(MonitorUpdate);

	bool dipsw1[8] {};
	bool dipsw2[8] {};
	bool int1rq {};			// $01 INT1 割り込み要求
	bool int1en {};			// $04 INT1 割り込み許可
	bool int5rq {};			// $08 INT5 割り込み要求
	bool int5en {};			// $10 INT5 割り込み許可
	bool parity {};			// $40 パリティ有効/無効
	bool xp_reset {};		// $80 XP リセット

	Monitor monitor { this };
};

class PIO1Device : public i8255Device
{
	using inherited = i8255Device;
 public:
	PIO1Device();
	virtual ~PIO1Device() override;

	void ResetHard() override;

 protected:
	// BusIO インタフェース
	static const uint32 NPORT = 4;
	uint64 Read(uint32 offset);
	uint64 Write(uint32 offset, uint32 data);
	uint64 Peek(uint32 offset);

 private:
	void SetPC(int, int) override;
	void PowerOffCallback(Event& ev);

	bool power {};			// パワーダウン (false で電源オフ指示)
	bool xp_intreq {};		// XP 割り込み要求。あり(true)なら %0

	// 電源オフイベント
	Event poffevent { this };
};

extern std::unique_ptr<PPIDevice> gPPI;
extern std::unique_ptr<PIO0Device> gPIO0;
extern std::unique_ptr<PIO1Device> gPIO1;
