//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// PROM (LUNA-I, LUNA88K 共通)

//          IODevice
//              |
//              +----------------------------------------+
//              |                                        |
//              v                                        v
//          ROMDevice                            Luna*PROMEmuDeivce
//              |                                (ROM エミュレーションはここ)
//    +---------+---------------+
//    v                         v
//  PROMDevice             CGROMDevice など
//  (LUNA* の実ROM共通)    (X68k の各種ROM)

#include "prom.h"
#include "config.h"
#include "mainapp.h"
#include "scheduler.h"

// コンストラクタ
PROMDevice::PROMDevice()
	: inherited("PROM")
{
	devaddr = 0x41000000;
}

// デストラクタ
PROMDevice::~PROMDevice()
{
}

bool
PROMDevice::Init()
{
	vmtype_t vmtype = gMainApp.GetVMType();

	switch (vmtype) {
	 case VMTYPE_LUNA1:
		return InitLuna1();

	 case VMTYPE_LUNA88K:
		return InitLuna88k();

	 default:
		PANIC("unknown vmtype %d", vmtype);
	}
}

// LUNA-I の場合
bool
PROMDevice::InitLuna1()
{
	devlen = 128 * 1024;
	mask = devlen - 1;

	// ファイル名が存在してる時しかここに来ない
	const ConfigItem& item = gConfig->Find("prom-image");
	if (LoadROM(item.AsString().c_str(), devlen) == false) {
		return false;
	}

	// ROM バージョンを確認。確認されてるのは
	// V4.02 (Wed Oct 12 17:07:11 1988)
	// V4.22 (Thu Jul 27 11:45:42 1989)
	// V4.25
	//
	// とりあえず V4.22 は +$00d390 からにタイムスタンプ文字列があるので
	// それで判別してみる。
	//00d390 54 68 75 20 4a 75 6c 20 32 37 20 31 31 3a 34 35 |Thu Jul 27 11:45|
	//00d3a0 3a 34 32 20 31 39 38 39 00 00 00 00 4e 56 fd f8 |:42 1989....NV..|

	if (strcmp((const char *)&mem[0xd390], "Thu Jul 27 11:45:42 1989") == 0) {
		// V4.22
		putlog(1, "V4.22 detected");

		// MAC アドレスは ROM に書き込まれているので、ここで上書き
		if (strcmp((const char *)&mem[0x1ffd8], "ENADDR") != 0) {
			warnx("ENADDR not found; unknown ROM V4.22 variants?");
			return false;
		}

		// MAC アドレスは12桁の16進数文字列として書き込まれている。例えば
		// 前半の 00:00:0A 部分は 30 30 30 30 30 41 "00000A" という感じ。
		const macaddr_t macaddr = gEthernet->GetMacAddr();
		memcpy(&mem[0x1ffe0], macaddr.to_string().c_str(), 12);

		// SCSI デバッグビット
		bool scsi_debug_bit = gConfig->Find(".prom-scsi-debug").AsInt();
		mem[0x4100981d - 0x41000000] = (scsi_debug_bit) ? 0x01 : 0;
	} else {
		// 他の ROM は何も分からないので何もしない。
		putlog(0, "Unknown PROM Version");
	}

	return true;
}

// LUNA88K の場合
bool
PROMDevice::InitLuna88k()
{
	devlen = 256 * 1024;
	mask = devlen - 1;

	// ファイル名が存在してる時しかここに来ない
	const ConfigItem& item = gConfig->Find("prom-image");
	if (LoadROM(item.AsString().c_str(), devlen) == false) {
		return false;
	}

	return true;
}

bool
PROMDevice::PowerOn()
{
	// パフォーマンス測定に電源オン時の実時刻を記憶。
	// Write8() 参照。
	sw.Start();
	return true;
}

void
PROMDevice::ResetHard()
{
	// PROM への Write (RAM への切り替え) は何度も実行されるので、最初の
	// 1回だけ処理するため、こちらでも状態を持つ。
	// Scheduler も同様のステートを持っているけど、あっちはリクエストと
	// 現状の2つがあって、リクエスト投げて状態が変わるまでに遅延があるので
	// たぶんこっちで処理したほうがいい。
	isrom = true;
	gVM->SwitchBootPage(this, true);
}

uint64
PROMDevice::Read8(uint32 addr)
{
	// パフォーマンス測定のためのバックドア。
	// 0x42000004 番地をバイト読み込みを行うと VM 電源オン時から現在までの
	// 実時間をホスト (ログ) に出力する。NetBSD のマルチユーザブートが
	// あらかた終わってログインプロンプトが出る少し前に実行される
	// /etc/rc.local に
	// dd if=/dev/mem skip=1107296260 bs=1 count=1
	// と書いておくことでここまでにかかる時間を表示できる。ただしこの値は
	// ホスト性能に依存するため、同一環境下での相対的な比較のみが可能。
	// 実機ではこの動作では PROM のミラーが読めるだけなので副作用はないはず。
	// PROM は仕様では 0x4100_0000 から 128KB なので、ここでは通常アクセス
	// するはずのない 0x4200_0004 からの1バイトの読み込みをバックドアに使う。
	// 0x4100_0000..0x41ff_ffff への書き込みは 0番地付近を ROM から RAM に
	// 切り替えるらしいが実際に PROM がアクセスするのは 0x4100_0000.[BL]
	// のみなので、ここではそれを避けて 0x4200_0004.B を使う。
	// 書き込みをバックドアにしたかったが dd で書き込めなさそうだったので
	// 読み込みを使う。ちゃんと調べてない。
	if (__predict_false(addr == 0x42000004)) {
		uint64 rtime;
		uint64 vtime;

		// 電源オン時からここまでの時間を表示するだけでリセット等はしない。
		sw.Stop();
		rtime = sw.Elapsed();
		vtime = gMPU->GetVirtTime();
		putlog(0, "電源オンから %d.%03d 秒経過 %3.3f%%",
			(int)(rtime / (1000 * 1000 * 1000)),
			(int)((rtime / (1000 * 1000)) % 1000),
			(double)vtime / rtime * 100);
	}

	// それ以外は通常動作
	return inherited::Read8(addr);
}

uint64
PROMDevice::Write8(uint32 addr, uint32 data)
{
	return Write(addr);
}

uint64
PROMDevice::Write16(uint32 addr, uint32 data)
{
	return Write(addr);
}

uint64
PROMDevice::Write32(uint32 addr, uint32 data)
{
	return Write(addr);
}

// Write{8,16,32} 共通。
// LUNA では書き込みでバスエラーは起きない。
// 電源オン後最初の書き込みでブートページを RAM に切り替える。
uint64
PROMDevice::Write(uint32 addr)
{
	if (__predict_false(isrom)) {
		// LUNA では PROM への書き込みでブートページを RAM に切り替えている。
		// - LUNA-I では最初に $41000000.B への書き込みがあるが、
		//   それ以降も $41000000.L への書き込みが(複数回)あるようだ。
		// - LUNA88K では $4100000c.L への書き込みがある。
		// よってアドレスを定めず PROM 全域への書き込みで最初の1回だけ発動
		// するようにしておく。サイズも今の所ワード(16bit)アクセスは確認して
		// ないけど統一的に対処しておく。

		// XXX PROM がブートページを担当してる期間に RAM (のはずのアドレス)
		// への書き込みは起きるんだろうか?
		if (addr < 0x41000000) {
			putlogn("%s addr=$%08x", __func__, addr);
		}

		// RAM に切り替え
		isrom = false;
		gVM->SwitchBootPage(this, false);
	}

	// XXX ウェイト未調査
	gMPU->AddCycle(wait);

	// バスエラーは起きない
	return 0;
}
