//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "renderer.h"
#include "bt454.h"
#include "lunafb.h"
#include "mythread.h"
#include "tvram.h"

std::unique_ptr<Renderer> gRenderer;

static void *renderer_run(void *);

// コンストラクタ
Renderer::Renderer()
	: inherited("Renderer")
{
	// 初回は必ず描画する? (電源オン直後かな?)
	invalidate = true;
}

// デストラクタ
Renderer::~Renderer()
{
	// imagebuf の所有権は持っていないので解放不要。

	if (thread_created) {
		Terminate();
		pthread_join(thread, NULL);
	}
}

bool
Renderer::Init()
{
	// レンダリングスレッド起動
	pthread_create(&thread, NULL, renderer_run, NULL);
	thread_created = true;

	return true;
}

bool
Renderer::PowerOff()
{
	if (imagebuf) {
		// レンダリングスレッドに指示
		std::unique_lock<std::mutex> lock(mtx);
		request |= REQ_REND_POWEROFF;
		cv.notify_one();
	}

	return true;
}

// バッファとコールバックを指定する。
// 引数 buf と callback はどちらも NULL であってはならない。
// これは画面を必要とする wx 側から呼ばれる。CLI では呼ばれない。
void
Renderer::SetBuf(uint8 *buf, void (*callback)(void))
{
	assert(buf);
	assert(callback);

	imagebuf = buf;
	refresh_callback = callback;
}

// 画面合成処理 (CRTC とかから呼ばれる部分)
// 今は一画面単位で合成を行なっている。ラスター云々は未対応。
// 画面合成を行うかどうかとか、レンダリングスレッドに指示を出すかとか
// どこら辺までを呼び出し元とこっちが分担するかは未定。
void
Renderer::Render()
{
	// CLI 版なら何もしない。
	if (imagebuf == NULL) {
		return;
	}

	// スケジューラ都合による間引きとかがあればここらへんで評価?
	// XXX どっちのスレッドでやるか

	// レンダリングスレッドに指示
	std::unique_lock<std::mutex> lock(mtx);
	request |= REQ_REND_NORMAL;
	cv.notify_one();
}

// 全画面無効化の指示 (パレット変更で呼ばれる)。
// ここでは無効化だけ受け付けておいて、実際の処理はレンダリングの際に行う。
void
Renderer::Invalidate()
{
	std::unique_lock<std::mutex> lock(mtx);
	request |= REQ_INVALIDATE;
	cv.notify_one();
}

// レンダリングスレッドのエントリポイント
void *
renderer_run(void *dummy)
{
	PTHREAD_SETNAME("Renderer");

	gRenderer->ThreadRun();
	return NULL;
}

// レンダリングスレッド
void
Renderer::ThreadRun()
{
	for (;;) {
		uint32 req;

		// 何か起きるまで待つ
		{
			std::unique_lock<std::mutex> lock(mtx);
			cv.wait(lock, [&] { return (request != 0); });
			req = request;
			request = 0;
		}

		if ((req & REQ_TERMINATE)) {
			// 終了するので、他のフラグは無視
			break;
		}
		if ((req & REQ_REND_POWEROFF)) {
			// 電源オフ時の描画 (通常描画フラグは無視)

			// 電源オフされると画面を灰色にする。
			// 黒でもいいけど、電源オンで黒いままなのと区別したい。
			memset(imagebuf, 0x70, width * height * BPP);
			// ウィンドウに通知
			refresh_callback();
			continue;
		}
		if ((req & REQ_INVALIDATE)) {
			// 全画面無効化の指示。ここではフラグを立てるだけ。
			// パレットは普段はまず変更されないが、変更する時には連続して
			// 全パレットが変更されがちなので、ここでフラグを蓄積する。
			// このフラグは同じスレッドの RenderMain() 内で回収する。
			invalidate = true;
		}
		if ((req & REQ_REND_NORMAL)) {
			// 通常描画

			// レンダラを駆動して
			if (RenderMain()) {
				// 更新があればウィンドウに通知
				refresh_callback();
			}
		}
	}
}

// スレッドの終了を指示 (スレッドは動いていること)
void
Renderer::Terminate()
{
	assert(thread_created);

	std::unique_lock<std::mutex> lock(mtx);
	request |= REQ_TERMINATE;
	cv.notify_one();
}

//
// X68030 レンダラ
//

// コンストラクタ
X68030Renderer::X68030Renderer()
{
	// XXX とりあえず
	width = 768;
	height = 512;
}

// デストラクタ
X68030Renderer::~X68030Renderer()
{
}

// 画面合成
bool
X68030Renderer::RenderMain()
{
	assert(imagebuf);

	// 今はテキスト画面のみを合成バッファに直接描いている
	return gTVRAM->Render(imagebuf);
}

//
// LUNA レンダラ
//

// コンストラクタ
LunaRenderer::LunaRenderer()
{
	width = 1280;
	height = 1024;
}

// デストラクタ
LunaRenderer::~LunaRenderer()
{
}

bool
LunaRenderer::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	// ホストパレット配列を取得。
	pal = gBT454->GetHostPalette();
	return true;
}


// 画面合成
bool
LunaRenderer::RenderMain()
{
	const uint32 *plane0;
	std::atomic<uint8> *dirty;
	int hscroll;
	int vscroll;
	uint8 *dst0;
	bool updated;

	assert(imagebuf);
	dst0 = imagebuf;

	// ビットマップ
	gLunafb->GetRFCNT(&hscroll, &vscroll);
	const uint8 *mem = gLunafb->GetBuf();
	dirty = gLunafb->GetDirty();

	updated = false;

	// dirty は VRAM への書き込みがあれば常に 1 にする。
	// レンダリング時に dirty が 1 なら 2 にしてレンダリング。
	// レンダリング終了時に dirty が 2 なら 0 に戻す。
	// これによりレンダリング中に VRAM への書き込みがあれば次回描画される。

	// 全域を無効化して画面全体を描画し直す。
	// XXX この処理中に VRAM 書き込み等があっても整合するんだっけこれ
	if (invalidate) {
		for (int y = 0; y < height; y++) {
			dirty[y] = 1;
		}
		invalidate = false;
	}

	// sy はメモリ上の Y 座標
	// dy はレンダラ画面の Y 座標 (RFCNT のラスタカウンタを考慮)

	// XXX とりあえず1プレーン
	plane0 = (const uint32 *)mem;
	int sy = vscroll;
	for (int dy = 0; dy < height; dst0 += width * BPP, dy++, sy++) {
		sy &= 0x3ff;

		if (atomic_cas_8(&dirty[sy], 1, 2) == 1) {
			uint8 *dst = dst0;
			// sx はメモリ上の X オフセット (ロングワード単位)
			// dx はレンダラ画面の X オフセット (ロングワード単位)
			// XXX 水平オフセットが仕様以上の時の挙動は未実装

			int sx = (sy * 2048 + hscroll) / 32;
			int shift = (hscroll & 31);
			for (int dx = 0; dx < width / 32; dx++, sx++) {
				uint32 p0 = plane0[sx];
				if (shift) {
					p0 = (p0 << shift) | (plane0[sx + 1] >> (32 - shift));
				}

				// 1ロングワードの横 32bit が横 32dot に相当
				for (int j = 0; j < 32; j++) {
					// 1bpp の時 P1,P2,P3 分はプルアップされて %1 に
					// なっているっぽいので、実質 #14 か #15 ということ。
					uint c = 14;
					// カラーコード(パレット番号)を作成
					if ((p0 & 0x80000000U)) c |= 1;
					p0 <<= 1;

					*dst++ = pal[c * 4 + 0];
					*dst++ = pal[c * 4 + 1];
					*dst++ = pal[c * 4 + 2];
				}
			}

			// 2の時だけ0に戻す
			atomic_cas_8(&dirty[sy], 2, 0);
			updated = true;
		}
	}

	return updated;
}
