//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "rp5c15.h"
#include "mpu.h"
#include <sys/time.h>

// アラームは未実装

// XXX RTC の読み書きはロックしたほうがいいような気がする

RP5C15Device::RP5C15Device()
	: inherited("RTC")
{
	devaddr = 0xe8a000;
	devlen  = 0x2000;

	monitor.func = (MonitorCallback_t)&RP5C15Device::MonitorUpdate;
	monitor.SetSize(30, 17);
	monitor.Regist(ID_MONITOR_RTC);
}

RP5C15Device::~RP5C15Device()
{
}

// 初期化
bool
RP5C15Device::Init()
{
	// ホストの時刻をローカル時刻で取得
	LoadHostRTC(true);

	// それをレジスタに反映。
	UpdateReg(UPDATE_ALL);

	return true;
}

uint64
RP5C15Device::Read(uint32 offset)
{
	uint8 data;
	int n;

	gMPU->AddCycle(29);	// InsideOut p.135

	// バンクを展開する
	n = bank * 0x10 + offset;

	// XXX 14, 15 は write only らしいが何が読めるか

	switch (n) {
	 case 0x1d:	// MODE  (Bank0 と同じ)
	 case 0x1e:	// TEST  (Bank0 と同じ)
	 case 0x1f:	// RESET (Bank0 と同じ)
		n -= 0x10;
		FALLTHROUGH;

	 case 0x00 ... 0x1c:
		data = reg.r[n];
		break;
	 default:
		__unreachable();
	}

	// 下位4ビット(RP5C15) の中の未実装ビットは %0 だが
	// 上位4ビットは %1 になる。
	data |= 0xf0;
	putlog(2, "RTC Bank%d $%x -> $%02x", bank, offset, data);

	return data;
}

uint64
RP5C15Device::Write(uint32 offset, uint32 data)
{
	int n;

	gMPU->AddCycle(38);	// InsideOut p.135

	// バンクを展開する
	n = bank * 0x10 + offset;

	putlog(2, "RTC Bank%d $%x <- $%x", bank, offset, (data & 0x0f));

	switch (n) {
	 case RP5C15::SEC1:
		rtc.sec = reg.sec10 * 10 + (data & reg.mask[n]);
		UpdateReg(UPDATE_SEC);
		break;
	 case RP5C15::SEC10:
		rtc.sec = (data & reg.mask[n]) * 10 + reg.sec1;
		UpdateReg(UPDATE_SEC);
		break;

	 case RP5C15::MIN1:
		rtc.min = reg.min10 * 10 + (data & reg.mask[n]);
		UpdateReg(UPDATE_MIN);
		break;
	 case RP5C15::MIN10:
		rtc.min = (data & reg.mask[n]) * 10 + reg.min1;
		UpdateReg(UPDATE_MIN);
		break;

	 case RP5C15::HOUR1:	// hour1
		if (reg.sel24) {
			// 24時間計
			rtc.hour = reg.hour10 * 10 + (data & reg.mask[n]);
		} else {
			// 12時間計
			rtc.hour = ((reg.hour10 & 0x02) ? 12 : 0)
			          + (reg.hour10 & 0x01) * 10 + (data & reg.mask[n]);
		}
		UpdateReg(UPDATE_HOUR);
		break;
	 case RP5C15::HOUR10:	// hour10
		if (reg.sel24) {
			// 24時間計
			rtc.hour = (data & reg.mask[n]) * 10 + reg.hour1;
		} else {
			// 12時間計
			rtc.hour = ((data & 0x02) ? 12 : 0)
			          + (data & 0x01) * 10 + reg.hour1;
		}
		UpdateReg(UPDATE_HOUR);
		break;

	 case RP5C15::WDAY:
		// 曜日カウンタは日付などとは一切連動していない
		rtc.wday = data & reg.mask[n];
		UpdateReg(UPDATE_DAY);
		break;

	 case RP5C15::DAY1:
		rtc.mday = reg.day10 * 10 + (data & reg.mask[n]);
		UpdateReg(UPDATE_DAY);
		break;
	 case RP5C15::DAY10:
		rtc.mday = (data & reg.mask[n]) * 10 + reg.day1;
		UpdateReg(UPDATE_DAY);
		break;

	 case RP5C15::MON1:
		rtc.mon = reg.mon10 * 10 + (data & reg.mask[n]);
		UpdateReg(UPDATE_MON);
		break;
	 case RP5C15::MON10:
		rtc.mon = (data & reg.mask[n]) * 10 + reg.mon1;
		UpdateReg(UPDATE_MON);
		break;

	 case RP5C15::YEAR1:
		rtc.year = 1980 + (reg.year10 * 10) + (data & reg.mask[n]);
		UpdateReg(UPDATE_YEAR);
		break;
	 case RP5C15::YEAR10:
		rtc.year = 1980 + (data & reg.mask[n]) * 10 + reg.year1;
		UpdateReg(UPDATE_YEAR);
		break;

	 case RP5C15::CLKSEL:
	 case 0x12 ... 0x1c:
		// マスクして書き込む
		reg.r[n] = data & reg.mask[n];
		break;

	 case RP5C15::ADJUST:
	 {
		// %0 なら何もしない?
		if ((data & 1) == 0) {
			break;
		}
		// %1 ならアジャスト。

		// 現在の秒が 0..29秒なら 0秒に切り下げ。
		// 30..59秒なら +1分0秒に繰り上げ。
		// いずれにしても 1秒と10秒の位は 0 にする。
		uint32 updates = UPDATE_SEC;
		if (rtc.sec < 30) {
			// 1分切り下げ
			rtc.sec = 0;
		} else {
			// 1分繰り上げ
			rtc.sec = 0;
			updates = CarryMin(updates);
		}
		UpdateReg(updates);
		break;
	 }

	 case RP5C15::MODE:
	 case RP5C15::MODE_BANK1:
		reg.mode = data & reg.mask[RP5C15::MODE];
		// bank を更新
		bank = (data & RP5C15::MODE_BANK);
		break;

	 case RP5C15::TEST:
	 case RP5C15::TEST_BANK1:
		data &= reg.mask[RP5C15::TEST];
		if (data == 0) {
			// 通常 0 を書き込むこと、となっているので 0 は許可
		} else {
			putlog(0, "TEST <- $%1x 未実装書き込み", data);
		}
		reg.test = data;
		break;

	 case RP5C15::RESET:
	 case RP5C15::RESET_BANK1:
		reg.reset = data & reg.mask[RP5C15::RESET];
		if ((reg.reset & RP5C15::RESET_TIMER)) {
			// %1 を書き込むと、秒未満の内部カウンタをリセット
			// 繰り上がり分の保持も消えるけどいいよな?
			rtc.cnt = 0;
		}
		break;

	 default:
		break;
	}
	return 0;
}

uint64
RP5C15Device::Peek(uint32 offset)
{
	int n;

	// バンクを展開する
	n = bank * 0x10 + offset;

	switch (n) {
	 case RP5C15::MODE_BANK1:
	 case RP5C15::TEST_BANK1:
	 case RP5C15::RESET_BANK1:
		// Bank0 と同じものが見える
		n -= 0x10;
		FALLTHROUGH;

	 case 0x00 ... 0x1c:
		return reg.r[n] | 0xf0;

	 default:
		__unreachable();
	}
}

void
RP5C15Device::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int x;
	int y;

	//
	//  #0 #1
	// 1:4 2
	//
	//        0 1 2 3 4 5 6 7 8 9 A B C D E F
	// Bank0  0 1 2 3 4 5 6 7 8 9 a b c d e f
	// Bank1

	screen.Clear();

	screen.Puts(0, 0, " #0 #1");
	y = 1;
	for (int i = 0; i < 13; i++, y++) {
		screen.Print(0, y, "%X:%x %x", i, reg.r[i], reg.r[i + 16]);
	}
	for (int i = 13; i < 16; i++, y++) {
		screen.Print(0, y, "%X:%x", i, reg.r[i]);
	}

	x = 8;
	y = 0;

	int year = reg.year10 * 10 + reg.year1 + 1980;
	int mon  = reg.mon10 * 10 + reg.mon1;
	int day  = reg.day10 * 10 + reg.day1;
	screen.Print(x, y, "%04d/%02d/%02d(%s)",
		year, mon, day, wdays[reg.wday]);
	y++;

	int hour;
	if (reg.sel24) {
		hour = reg.hour10 * 10 + reg.hour1;
	} else {
		hour = (reg.hour10 & 1) * 10 + reg.hour1;
		screen.Puts(x, y, (reg.hour10 & 2) == 0 ? "AM" : "PM");
	}
	int min = reg.min10 * 10 + reg.min1;
	int sec = reg.sec10 * 10 + reg.sec1;
	screen.Print(x + 3, y, "%02d:%02d:%02d", hour, min, sec);
	y++;

	screen.Print(x, y++, "Leap %d", reg.leap);
	y++;

	screen.Print(x, y++, "CLKOUT %c", clkout ? 'H' : 'L');
	screen.Print(x, y++, "ALARM  %c", alarmout ? 'H' : 'L');
	y++;

	screen.Print(x, y++, "cnt  %2d", rtc.cnt);
}

// 32Hz クロック入力
//
// 32Hz ごとに呼ばれて、これが RTC の時計を進める処理をするところ。
// RTC を実時間に同期して進めるモードなら実時間を計るスケジューラから、
// RTC を仮想時間に同期して進めるモードなら VM スケジューラから呼ばれる。
//
// CLKOUT 128Hz 以上のモードには対応していない。
void
RP5C15Device::ClockIn()
{
	uint32 updates;

	// 32Hz カウンタを進める
	updates = CountUp();

	// レジスタの値を更新する
	UpdateReg(updates);

	// CLKOUT 端子の切り替え周期は CLKSEL による
	switch (reg.clksel) {
	 case 0:	// ハイインピーダンス
	 case 1:	// 16384Hz
	 case 2:	// 1024Hz
	 case 3:	// 128Hz
		// 128Hz 点滅とか点灯みたいなもんだろ(適当)
		clkout = true;
		break;

	 case 4:	// 16Hz
		clkout = !clkout;
		// アラーム端子 16Hz パルス出力は %0 ならオン
		if ((reg.reset & RP5C15::RESET_16Hz) == 0) {
			alarmout = !alarmout;
		}
		break;

	 case 5:	// 1Hz
		// CLKOUT 端子は秒の繰り上がりに同期している。
		// ALARM 端子の 1Hz パルスもそれに同期するとは書いてないけど、
		// 面倒なのでとりあえず同じところで処理してしまう。
		// TIMER_EN (秒への繰り上がり) が無効でも反転自体は起きる気が
		// するので、ここでは UPDATE_SEC ではなく cnt で判定する。
		if ((rtc.cnt & 31) == 0) {
			// CLKOUT は 1秒の繰り上がりが起きる時に 'H' になる
			clkout = true;

			// ALARM 反転
			if ((reg.reset & RP5C15::RESET_1Hz) == 0) {
				alarmout = !alarmout;
			}
		} else if ((rtc.cnt & 31) == 16) {
			// CLKOUT、その反対側
			clkout = false;

			// ALARM 反転
			if ((reg.reset & RP5C15::RESET_1Hz) == 0) {
				alarmout = !alarmout;
			}
		}
		break;

	 case 6:	// 1/60Hz
		if ((updates & UPDATE_MIN)) {
			// 00秒になったので CLKOUT を 'H' に
			clkout = true;
		} else if ((updates & UPDATE_SEC) && rtc.sec == 30) {
			// 30秒になったので CLKOUT を 'L' に
			clkout = false;
		}
		break;

	 case 7:	// 'L' 固定
		clkout = false;
		break;
	}
}

// updates で指定された桁に該当するレジスタを更新する。
// 繰り上がりかどうかは処理せず現在の値を RP5C15 レジスタにセットするだけ。
void
RP5C15Device::UpdateReg(uint32 updates)
{
	if ((updates & UPDATE_SEC)) {
		reg.sec1  = rtc.sec % 10;
		reg.sec10 = rtc.sec / 10;
	}
	if ((updates & UPDATE_MIN)) {
		reg.min1  = rtc.min % 10;
		reg.min10 = rtc.min / 10;
	}
	if ((updates & UPDATE_HOUR)) {
		if (reg.sel24) {
			// 24時間計モードでは普通の BCD
			reg.hour1  = rtc.hour % 10;
			reg.hour10 = rtc.hour / 10;
		} else {
			// 12時間計モードなら、時のbit1 が AM/PM を表す
			int h12 = rtc.hour % 12;
			reg.hour1  = h12 % 10;
			reg.hour10 = h12 / 10;
			reg.hour10 |= (rtc.hour >= 12) ? 0x02 : 0x00;
		}
	}

	if ((updates & UPDATE_DAY)) {
		reg.wday  = rtc.wday;
		reg.day1  = rtc.mday % 10;
		reg.day10 = rtc.mday / 10;
	}
	if ((updates & UPDATE_MON)) {
		reg.mon1  = rtc.mon % 10;
		reg.mon10 = rtc.mon / 10;
	}
	if ((updates & UPDATE_YEAR)) {
		// X680x0 では 0 を 1980年としている。
		int y = rtc.year - 1980;
		reg.year1  = y % 10;
		reg.year10 = y / 10;
	}
}

/*static*/ const uint8
RP5C15::mask[] = {
	0x0f,	0x07,	0x0f,	0x07,	// 1秒	10秒	1分		10分
	0x0f,	0x03,	0x07,	0x0f,	// 1時	10時	曜日	1日
	0x01,	0x0f,	0x01,	0x0f,	// 10日	1月		10月	1年
	0x0f,	0x0d,	0x0f,	0x0f,	// 10年	MODE	TEST	RESET

	0x07,	0x01,	0x0f,	0x07,	// CLKO	Adjust	AL1分	AL10分
	0x0f,	0x03,	0x07,	0x0f,	// AL1時 AL10時	AL曜日	AL1日
	0x03,	0x00,	0x01,	0x03,	// AL10日 ---	12/24	閏年
	0x00							// ---
};
