//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "device.h"

// このクラスは SCSI プロトコル上の定数などだけを持つ。
class SCSI
{
 public:
	static const uint IO		= 0x01;
	static const uint CD		= 0x02;
	static const uint MSG		= 0x04;
	static const uint BSY		= 0x08;
	static const uint SEL		= 0x10;

	// SCSI のフェーズを表す。
	// フェーズは概ね BSY と SEL の状態遷移によって決まる。
	// 情報転送フェーズの内訳は XferPhase 側で区別する。
	enum Phase {
		BusFree		= 0,
		Arbitration	= 1,
		Selection	= 2,
		Reselection	= 3,
		Transfer	= 4,
	};

	// 情報転送フェーズを表す。
	// この値は MSG,CD,IO で表現できる値をそのまま流用して表現する。
	enum XferPhase {    	// MSG CD IO
		DataOut		= 0,	//   0  0  0
		DataIn		= 1,	//   0  0  1
		Command		= 2,	//   0  1  0
		Status		= 3,	//   0  1  1
		MsgOut		= 6,	//   1  1  0
		MsgIn		= 7,	//   1  1  1

		End			= -1,	// SCSICmd 内でフェーズ遷移を表現するために使う
	};

	// フェーズ文字列を返す
	static const char * GetPhaseName(int phase) {
		return GetPhaseName((Phase)phase);
	}
	static const char * GetPhaseName(Phase phase) {
		switch (phase) {
		 case BusFree:		return "BusFree";
		 case Arbitration:	return "Arbitration";
		 case Selection:	return "Selection";
		 case Reselection:	return "Reselection";
		 case Transfer:		return "Transfer";
		}
		return "Phase?";
	}

	// 情報転送フェーズ文字列を返す
	static const char * GetXferPhaseName(int xfer) {
		return GetXferPhaseName((XferPhase)xfer);
	}
	static const char * GetXferPhaseName(XferPhase xfer) {
		switch (xfer) {
		 case DataOut:	return "DataOut";
		 case DataIn:	return "DataIn";
		 case Command:	return "Command";
		 case Status:	return "Status";
		 case MsgOut:	return "MsgOut";
		 case MsgIn:	return "MsgIn";
		 case End:		return "End";
		}
		return "XferPhase?";
	}

	// デバイス種別
	enum class DevType {
		None = 0,
		Initiator,
		HD,
	};

	// コマンド。
	// ダイレクトアクセスデバイスのコマンドを元にしている。
	// シーケンシャルアクセスデバイスでは 0x01 が Rewind だったり
	// 同じコードで違うコマンドのようだが、とりあえずそれは放置。
	struct Command {
		// Group0
		static const uint TestUnitReady		= 0x00;		// 共通
		static const uint RezeroUnit		= 0x01;
		static const uint RequestSense		= 0x03;		// 共通
		static const uint FormatUnit		= 0x04;
		static const uint ReassignBlocks	= 0x07;
		static const uint Read6				= 0x08;
		static const uint Write6			= 0x0a;
		static const uint Seek6				= 0x0b;
		static const uint Inquiry			= 0x12;		// 共通
		static const uint ModeSelect6		= 0x15;		// 共通
		static const uint Reserve			= 0x16;
		static const uint Release			= 0x17;
		static const uint Copy				= 0x18;		// 共通
		static const uint ModeSense6		= 0x1a;
		static const uint StartStopUnit		= 0x1b;
		static const uint ReceiveDiagnosticResults	= 0x1c;	// 共通
		static const uint SendDiagnostic	= 0x1d;		// 共通
		static const uint PreventAllowMediumRemoval	= 0x1e;
		// Group1
		static const uint ReadCapacity		= 0x25;
		static const uint Read10			= 0x28;
		static const uint Write10			= 0x2a;
		static const uint Seek10			= 0x2b;
		// Group1
		static const uint WriteAndVerify10	= 0x2e;
		static const uint Verify10			= 0x2f;
		static const uint SearchDataHigh10	= 0x30;
		static const uint SearchDataEqual10	= 0x31;
		static const uint SearchDataLow10	= 0x32;
		static const uint SetLimits			= 0x33;
		static const uint PreFetch			= 0x34;
		static const uint SynchronizeCache	= 0x35;
		static const uint LockUnlockCache	= 0x36;
		static const uint ReadDefectData10	= 0x37;
		static const uint Compare			= 0x39;		// 共通
		static const uint CopyAndVerify		= 0x3a;		// 共通
		static const uint WriteBuffer		= 0x3b;		// 共通
		static const uint ReadBuffer		= 0x3c;		// 共通
		static const uint ReadLong			= 0x3e;
		static const uint WriteLong			= 0x3f;
		// Group2
		static const uint ChangeDefinition	= 0x40;
		static const uint WriteSame			= 0x41;
		static const uint LogSelect			= 0x4c;		// 共通
		static const uint LogSense			= 0x4d;		// 共通
		static const uint ModeSelect10		= 0x55;		// 共通
		static const uint ModeSense10		= 0x5a;		// 共通
		static const uint MAX				= 0x5b;	// 最大値+1にすること
	};

	// ステータスバイト
	struct StatusByte {
		static const uint8 Good					= (0x00 << 1);
		static const uint8 CheckCondition		= (0x01 << 1);
		static const uint8 ConditionMet			= (0x02 << 1);
		static const uint8 Busy					= (0x04	<< 1);
		static const uint8 Intermediate			= (0x08 << 1);
		static const uint8 IntermediateCondMet	= (0x0a << 1);
		static const uint8 ReservationConflict	= (0x0c << 1);
		static const uint8 CommandTerminated	= (0x11 << 1);
		static const uint8 QueueFull			= (0x14 << 1);
	};

	// メッセージバイト
	struct MsgByte {
		static const uint8 CommandComplete			= 0x00;
		static const uint8 ExtendMessage			= 0x01;
		static const uint8 SaveDataPointer			= 0x02;
		static const uint8 RestorePointers			= 0x03;
		static const uint8 Disconnect				= 0x04;
		static const uint8 InitiatorDetectedError	= 0x05;
		static const uint8 Abort					= 0x06;
		static const uint8 MessageReject			= 0x07;
		static const uint8 NoOperation				= 0x08;
		static const uint8 MessageParityError		= 0x09;
		static const uint8 LinkedCommandComplete	= 0x0a;
		static const uint8 LinkedCommandCompleteWithFlag = 0x0b;
		static const uint8 BusDeviceReset			= 0x0c;
		static const uint8 AbortTag					= 0x0d;
		static const uint8 ClearQueue				= 0x0e;
		static const uint8 InitiateRecovery			= 0x0f;
		static const uint8 ReleaseRecovery			= 0x10;
		static const uint8 TerminateIOProcess		= 0x11;
		static const uint8 SimpleQueueTag			= 0x20;
		static const uint8 HeadOfQueueTag			= 0x21;
		static const uint8 OrderedQueueTag			= 0x22;
		static const uint8 IgnoreWideResidue		= 0x23;
		static const uint8 Identify					= 0x80;	// 0x80..0xff
	};

	// センスキー
	struct SenseKey {
		static const uint8 NoSense			= 0x0;
		static const uint8 RecoveredError	= 0x1;
		static const uint8 NotReady			= 0x2;
		static const uint8 MediumError		= 0x3;
		static const uint8 HardwareError	= 0x4;
		static const uint8 IllegalRequest	= 0x5;
		static const uint8 UnitAttention	= 0x6;
		static const uint8 DataProtect		= 0x7;
		static const uint8 BlankCheck		= 0x8;
		static const uint8 VendorSpecific	= 0x9;
		static const uint8 CopyAborted		= 0xa;
		static const uint8 AbortedCommand	= 0xb;
		static const uint8 Equal			= 0xc;
		static const uint8 VolumeOverflow	= 0xd;
		static const uint8 Miscompare		= 0xe;
		static const uint8 Reserved			= 0xf;
	};

	// ASC/ASCQ
	// 上位バイトが ASC、下位バイトが ASCQ を示す。
	struct ASC {
		static const uint16 NoAdditionalSenseInformation	= 0x0000;
		static const uint16 InvalidCommandOperationCode		= 0x2000;
		static const uint16 InvalidFieldInCDB				= 0x2400;
	};

	// Mode Select(6) ヘッダ
	struct ModeParamHdr {
		uint8 mode_param_len;		// バイト長 (このバイトを含まずこれ以降)
		uint8 media_type;
		uint8 device_specific_param;
		uint8 block_desc_len;		// ブロックディスクリプタ長
	} __packed;

	// Mode Select(6) ブロックディスクリプタ
	struct BlockDesc {
		uint8 density_code;
		uint8 block_cnt[3];			// ブロック数
		uint8 reserved;
		uint8 block_size[3];		// 論理ブロック長 (バイト)
	} __packed;

	// SCSI コマンド名を返す
	static const char *GetCommandName(uint8 cmd);

 private:
	static const char * const commandname[Command::MAX];
};
