//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// SCSI デバイス共通部分

#include "scsidev.h"
#include "config.h"
#include "mainapp.h"

//
// SCSI デバイス
//

// コンストラクタ
SCSIDevice::SCSIDevice(const std::string& objname_)
	: inherited(objname_)
{
}

// デストラクタ
SCSIDevice::~SCSIDevice()
{
}

//
// SCSI ホストデバイス
//

// コンストラクタ
SCSIHostDevice::SCSIHostDevice(const std::string& objname_)
	: inherited(objname_)
{
	devtype = SCSI::DevType::Initiator;

	// バスデバイスを作成する。
	// ただし自身がバスに参加するのは継承側で行う
	// (例えば SPC なら BDID 書き込みで ID が決まった後、など)。
	scsibus.reset(new SCSIBus());
}

// デストラクタ
SCSIHostDevice::~SCSIHostDevice()
{
}

// 設定ファイルに使われるこのホストデバイスを示す接頭語を設定する
// (継承クラス側のコンストラクタで config_name を設定後に呼ぶこと)
void
SCSIHostDevice::SetConfigName(const char *name_)
{
	config_name = name_;
}

bool
SCSIHostDevice::Create()
{
	// 設定を読み込む。
	for (int id = 0; id < 8; id++) {
		const std::string key = string_format("%s-id%d-image",
			GetConfigName(), id);
		const ConfigItem& item = gConfig->Find(key);
		const std::string& image = item.AsString();

		// 空ならデバイスなし
		if (image.length() == 0) {
			// 未接続部分のパラメータは減らしておくか。
			// --show-config するとさすがに無駄に多くて邪魔なので。
			const std::string wpkey = string_format("%s-id%d-writeprotect",
				GetConfigName(), id);
			const std::string stkey = string_format("%s-id%d-seektime",
				GetConfigName(), id);
			gConfig->Delete(wpkey);
			gConfig->Delete(stkey);
			continue;
		}

		// ',' より前がデバイス種別 (ここではパスは不要)
		auto pos = image.find(',');
		if (pos == std::string::npos) {
			item.Err("no device type specified");
			return false;
		}
		std::string type = string_tolower(image.substr(0, pos));

		if (type == "hd") {
			target[id].reset(new SCSIHD(this, id));
			scsibus->Attach(target[id].get(), id);
		} else {
			item.Err("invaild device type '%s'", type.c_str());
		}
	}

	return true;
}

bool
SCSIHostDevice::Init()
{
	// 初期化ではないけどログレベルがセットできたここでデバッグ表示。
	// -L オプションの処理は Create() の後なので。

	if (loglevel >= 1) {
		for (int id = 0; id < 8; id++) {
			std::string typestr;
			if (target[id]) {
				switch (target[id]->GetDevType()) {
				 default:
					break;
				 case SCSI::DevType::HD:
					typestr = "HD";
					break;
				}
			}
			if (!typestr.empty()) {
				putmsgn("%s target #%d %s", GetConfigName(), id,
					typestr.c_str());
			}
		}
	}

	return true;
}

// バスフリーコールバック (SCSIBus から呼ばれる)
void
SCSIHostDevice::BusFreeCB(int id)
{
	// 実際にはバスフリーを検出したデバイスは各自自身が送出している信号を
	// 下げるのだが、色々面倒なので、ここではホストのバスフリーコールバックが
	// 一括して全部の線を下げる。
	// ただしさすがにログがうるさいので、上がってる線だけ下げる。

	uint32 data;
	while ((data = GetBSY()) != 0) {
		bus->NegateBSY(DecodeID(data));
	}
	while ((data = GetSEL()) != 0) {
		bus->NegateSEL(DecodeID(data));
	}
	if (GetREQ()) {
		NegateREQ();
	}
	if (GetACK()) {
		NegateACK();
	}
	if (GetATN()) {
		NegateATN();
	}
	SetXfer(0);
	SetData(0);
}

#if defined(__clang__)
// デストラクタがインラインで書けないのは面倒なのでとりあえず
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wweak-vtables"
#endif

//
// SCSI コマンドの共通部分
//
class SCSICmd
{
 public:
	SCSICmd(SCSITarget *p) {
		parent = p;
	}
	virtual ~SCSICmd() { }

	// 転送バッファと転送バイト数。
	// IN (ターゲットが送出側)なら、各フェーズ開始までに送出バイト数分の
	// バッファを buf に用意すること。recvbytes は操作不要。
	//
	// OUT(ターゲットが受信側)なら、各フェーズ開始前までに受信バイト数を
	// recvbytes にセットすること。buf は操作不要。
	// なお parent は OUT(受信)フェーズ中にこの recvbytes を変更しない。
	std::vector<uint8> buf;
	uint32 recvbytes;

	// このコマンドの動作時間 [nsec]
	// シークタイムとかを簡易的に再現するため、次フェーズ遷移時の最初の
	// REQ アサートを引き延ばすために使う。
	// 主にコマンドフェーズで使うことになると思う。
	uint64 optime {};

	// 以下いずれも戻り値は遷移先のフェーズ。

	// コマンドフェーズを処理する。
	// コマンドが確定したところで呼ばれる。戻り値に遷移先フェーズを返すこと。
	// また、コマンド開始時に各自で ClearSense() を呼ぶこと。センスというか
	// CA 状態は次のコマンド実行でクリアされるが (ネクサスはとりあえず無視)、
	// RequestSense の場合は現在保持しているセンス情報をホストに返すので、
	// つまり RequestSense 以外のコマンドが全員各自でクリアする必要がある。
	virtual SCSI::XferPhase Command(std::vector<uint8>& cmdseq) {
		parent->ClearSense();
		return SCSI::XferPhase::Status;
	}

	// データインフェーズを完了する。
	virtual SCSI::XferPhase DoneDataIn() { return SCSI::XferPhase::Status; }

	// データアウトフェーズを完了する。
	virtual SCSI::XferPhase DoneDataOut() { return SCSI::XferPhase::Status; }

	// ステータスフェーズを開始する。
	void BeginStatus() {
		buf.clear();
		buf.push_back(status_byte);
	}

	// ステータスフェーズを完了する。
	virtual SCSI::XferPhase DoneStatus() { return SCSI::XferPhase::MsgIn; }

	// メッセージインフェーズを開始する。
	// デフォルトでは Command Complete を返す実装を用意しておく。
	virtual void BeginMsgIn() {
		buf.clear();
		buf.push_back((uint8)SCSI::MsgByte::CommandComplete);
	}

	// メッセージインフェーズを完了する。
	virtual SCSI::XferPhase DoneMsgIn() { return SCSI::XferPhase::End; }

	// メッセージアウトフェーズを開始する。
	virtual void BeginMsgOut() { }

	// メッセージアウトフェーズを完了する。
	virtual SCSI::XferPhase DoneMsgOut() { return SCSI::XferPhase::End; }

 protected:
	// このコマンドを実行しているターゲットデバイス。
	SCSITarget *parent;

	// ステータスバイト (デフォルトは 0 = Good)
	uint8 status_byte {};
};

//
// SCSI コマンド 0x00: TestUnitReady (共通)
//
class SCSICmdTestUnitReady : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdTestUnitReady(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdTestUnitReady() override { }
};

//
// SCSI コマンド 0x01: RezeroUnit (共通)
//
class SCSICmdRezeroUnit : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdRezeroUnit(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdRezeroUnit() override { }

	// とりあえず黙って成功するだけでいいか
};

//
// SCSI コマンド 0x03: RequestSense (共通)
//
class SCSICmdRequestSense : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdRequestSense(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdRequestSense() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		// 応答のセンスデータは SCSI-2 形式(最小18バイト) を実装。
		// 今の所、センスキー、ASC、ASCQ だけで返せる内容にだけ対応している。
		// これ以外のフィールドを返す必要が出来たらまた考えること。

		buf.clear();
		buf.resize(18);
		buf[0] = 0xf0;				// Valid、エラーコード 0x70
		buf[1]  = 0;				// セグメント番号(?)
		buf[2]  = parent->sensekey;	// センスキー (上位4ビットは今の所 0)
		buf[3]  = 0;				// Information
		buf[4]  = 0;				// Information
		buf[5]  = 0;				// Information
		buf[6]  = 0;				// Information
		buf[7]  = 0;				// 追加センスデータ長 0
		buf[8]  = 0;				// Command specific Information
		buf[9]  = 0;				// Command specific Information
		buf[10] = 0;				// Command specific Information
		buf[11] = 0;				// Command specific Information
		buf[12] = parent->senseasc >> 8;
		buf[13] = parent->senseasc & 0xff;
		buf[14] = 0;				// FRU
		buf[15] = 0;				// センスキー固有情報
		buf[16] = 0;				// センスキー固有情報
		buf[17] = 0;				// センスキー固有情報
		// 追加センスデータなし

		// それからバッファを指定の長さに変更
		buf.resize((uint)cmdseq[4]);

		// 転送した後にセンスデータをクリア
		parent->ClearSense();

		return SCSI::XferPhase::DataIn;
	}
};

//
// SCSI コマンド 0x08: Read(6) (共通)
//
class SCSICmdRead6 : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdRead6(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdRead6() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		SCSIHD *hdparent = dynamic_cast<SCSIHD*>(parent);
		uint32 lba;
		uint32 blks;

		// センスデータをクリア
		parent->ClearSense();

		lba  = (cmdseq[1] & 0x1f) << 16;
		lba |= cmdseq[2] << 8;
		lba |= cmdseq[3];
		blks = cmdseq[4];
		// blks == 0 は 256 ブロック転送を意味する。
		if (blks == 0) {
			blks = 256;
		}
		parent->putlogf(1, lstr("READ6 LBA=0x%x blks=0x%x", lba, blks));

		if (((uint64)lba + (uint64)blks) * parent->GetBlocksize() >
		    hdparent->GetSize())
		{
			parent->sensekey = SCSI::SenseKey::IllegalRequest;
			parent->senseasc = SCSI::ASC::InvalidFieldInCDB;
			status_byte = SCSI::StatusByte::CheckCondition;
			return SCSI::XferPhase::Status;
		}

		uint8 *start = hdparent->mem + lba * parent->GetBlocksize();
		uint8 *end = start + blks * parent->GetBlocksize();
		buf.assign(start, end);

		// 平均シークタイム
		optime = hdparent->seektime;

		return SCSI::XferPhase::DataIn;
	}
};

//
// SCSI コマンド 0x0a: Write(6) (共通)
//
class SCSICmdWrite6 : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdWrite6(SCSITarget *p) : inherited(p) {
		hdparent = dynamic_cast<SCSIHD*>(parent);
	}
	virtual ~SCSICmdWrite6() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		// センスデータをクリア
		parent->ClearSense();

		lba  = (cmdseq[1] & 0x1f) << 16;
		lba |= cmdseq[2] << 8;
		lba |= cmdseq[3];
		blks = cmdseq[4];
		// blks == 0 は 256 ブロック転送を意味する。
		if (blks == 0) {
			blks = 256;
		}

		if (((uint64)lba + (uint64)blks) * parent->GetBlocksize() >
		    hdparent->GetSize())
		{
			parent->sensekey = SCSI::SenseKey::IllegalRequest;
			parent->senseasc = SCSI::ASC::InvalidFieldInCDB;
			status_byte = SCSI::StatusByte::CheckCondition;
			return SCSI::XferPhase::Status;
		}

		// 平均シークタイム
		optime = hdparent->seektime;

		recvbytes = blks * parent->GetBlocksize();
		return SCSI::XferPhase::DataOut;
	}

	SCSI::XferPhase DoneDataOut() override {
		parent->putlogf(1, lstr("WRITE6 LBA=0x%x blks=0x%x", lba, blks));
		if (hdparent->writeprotect == false) {
			memcpy(hdparent->mem + (lba * parent->GetBlocksize()),
				buf.data(), buf.size());
		}
		return SCSI::XferPhase::Status;
	}

 private:
	SCSIHD *hdparent;
	uint32 lba;
	uint32 blks;
};

//
// SCSI コマンド 0x12: Inquiry (共通)
//
class SCSICmdInquiry : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdInquiry(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdInquiry() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		// センスデータをクリア
		parent->ClearSense();

		// 応答データはターゲットデバイスが持っている
		buf = parent->inquiry_data;

		// 本来の応答データ長か reqlen の短いほうまで。
		uint reqlen = (uint)cmdseq[4];
		if (reqlen < buf.size()) {
			buf.resize(reqlen);
		}
		return SCSI::XferPhase::DataIn;
	}
};

//
// SCSI コマンド 0x15 ModeSelect(6) (共通)
//
class SCSICmdModeSelect6 : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdModeSelect6(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdModeSelect6() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		// センスデータをクリア
		parent->ClearSense();

		recvbytes = cmdseq[4];
		return SCSI::XferPhase::DataOut;
	}

	SCSI::XferPhase DoneDataOut() override {
		// Mode Select(6) のデータは以下の構造で並んでいる。
		//  ヘッダ
		//  ブロックディスクリプタ#1
		//   :
		//  ブロックディスクリプタ#n
		//  パラメータページ #1
		//   :
		//  パラメータページ #m

		const SCSI::ModeParamHdr *hdr = (const SCSI::ModeParamHdr *)buf.data();

		if (hdr->block_desc_len > 0) {
			const SCSI::BlockDesc *desc =
				(const SCSI::BlockDesc *)(buf.data() + sizeof(*hdr));
			uint blocksize;
			blocksize  = (desc->block_size[0]) << 16;
			blocksize |= (desc->block_size[1]) << 8;
			blocksize |= (desc->block_size[2]);

			// ブロック長を書き戻す
			parent->SetBlocksize(blocksize);
		}
		return SCSI::XferPhase::Status;
	}
};

//
// SCSI コマンド 0x1a ModeSense(6) (共通)
//
class SCSICmdModeSense6 : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdModeSense6(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdModeSense6() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		uint8 pagecode = (cmdseq[2] & 0x3f);

		// センスデータをクリア
		parent->ClearSense();

		// 未実装 (XXX センスデータは適当)
		parent->putlogf(1, lstr("ModeSense(6) PageCode $%02x 未実装",
			pagecode));
		parent->sensekey = SCSI::SenseKey::IllegalRequest;
		parent->senseasc = SCSI::ASC::InvalidFieldInCDB;
		status_byte = SCSI::StatusByte::CheckCondition;

		return SCSI::XferPhase::Status;
	}
};

//
// SCSI コマンド 0x1b StartStopUnit (ダイレクトアクセスデバイス)
//
class SCSICmdStartStopUnit : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdStartStopUnit(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdStartStopUnit() override { }

	// スピンアップ/ダウンやメディアのロード/アンロードを指示する。
	// ので今の所エミュレーション上は特にすることはない。
	// XXX それっぽい時間待つくらいはしてもいいかも知れない

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		parent->putlogf(1, lstr("StartStopUnit"));
		return inherited::Command(cmdseq);
	}
};

//
// SCSI コマンド 0x25 ReadCapacity (ダイレクトアクセスデバイス)
//
class SCSICmdReadCapacity : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdReadCapacity(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdReadCapacity() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		SCSIHD *hdparent = dynamic_cast<SCSIHD*>(parent);

		// センスデータをクリア
		parent->ClearSense();

		// PMI=1 は未サポート
		if ((cmdseq[8] & 0x01) != 0) {
			parent->putlogf(0, lstr("SCSI ReadCapacity PMI=1 not supported"));
		}

		// 応答データは
		// +00.L 最終論理ブロックアドレス
		// +04.L ブロック長
		uint32 blksize = parent->GetBlocksize();
		uint32 lastblk = (hdparent->GetSize() / blksize) - 1;
		buf.resize(8);
		buf[0] = lastblk >> 24;
		buf[1] = lastblk >> 16;
		buf[2] = lastblk >> 8;
		buf[3] = lastblk;
		buf[4] = blksize >> 24;
		buf[5] = blksize >> 16;
		buf[6] = blksize >> 8;
		buf[7] = blksize;

		return SCSI::XferPhase::DataIn;
	}
};

//
// SCSI コマンド 0x28 Read(10) (ダイレクトアクセスデバイス)
//
class SCSICmdRead10 : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdRead10(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdRead10() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		SCSIHD *hdparent = dynamic_cast<SCSIHD*>(parent);
		uint32 lba;
		uint32 blks;

		// センスデータをクリア
		parent->ClearSense();

		if ((cmdseq[1] & 0x01)) {
			parent->putlogf(0, lstr("SCSI Read(10) RelAdr not supported"));
		}

		lba  = cmdseq[2] << 24;
		lba |= cmdseq[3] << 16;
		lba |= cmdseq[4] << 8;
		lba |= cmdseq[5];
		blks  = cmdseq[7] << 8;
		blks |= cmdseq[8];
		parent->putlogf(1, lstr("READ10 LBA=0x%x blks=0x%x", lba, blks));

		if (((uint64)lba + (uint64)blks) * parent->GetBlocksize() >
		    hdparent->GetSize())
		{
			parent->sensekey = SCSI::SenseKey::IllegalRequest;
			parent->senseasc = SCSI::ASC::InvalidFieldInCDB;
			status_byte = SCSI::StatusByte::CheckCondition;
			return SCSI::XferPhase::Status;
		}

		uint8 *start = hdparent->mem + lba * parent->GetBlocksize();
		uint8 *end = start + blks * parent->GetBlocksize();
		buf.assign(start, end);

		// 平均シークタイム
		optime = hdparent->seektime;

		return SCSI::XferPhase::DataIn;
	}
};

//
// SCSI コマンド 0x2a Write(10) (ダイレクトアクセスデバイス)
//
class SCSICmdWrite10 : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdWrite10(SCSITarget *p) : inherited(p) {
		hdparent = dynamic_cast<SCSIHD*>(parent);
	}
	virtual ~SCSICmdWrite10() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		// センスデータをクリア
		parent->ClearSense();

		lba  = cmdseq[2] << 24;
		lba |= cmdseq[3] << 16;
		lba |= cmdseq[4] << 8;
		lba |= cmdseq[5];
		blks  = cmdseq[7] << 8;
		blks |= cmdseq[8];

		if (((uint64)lba + (uint64)blks) * parent->GetBlocksize() >
			hdparent->GetSize())
		{
			parent->sensekey = SCSI::SenseKey::IllegalRequest;
			parent->senseasc = SCSI::ASC::InvalidFieldInCDB;
			status_byte = SCSI::StatusByte::CheckCondition;
			return SCSI::XferPhase::Status;
		}

		// 平均シークタイム
		optime = hdparent->seektime;

		recvbytes = blks * parent->GetBlocksize();
		return SCSI::XferPhase::DataOut;
	}

	SCSI::XferPhase DoneDataOut() override {
		parent->putlogf(1, lstr("WRITE10 LBA=0x%x blks=0x%x", lba, blks));
		if (hdparent->writeprotect == false) {
			memcpy(hdparent->mem + (lba * parent->GetBlocksize()),
				buf.data(), buf.size());
		}
		return SCSI::XferPhase::Status;
	}

 private:
	SCSIHD *hdparent;
	uint32 lba;
	uint32 blks;
};

//
// SCSI コマンド サポートしていないコマンド
//
class SCSICmdNotSupportedCommand : public SCSICmd
{
	using inherited = SCSICmd;
 public:
	SCSICmdNotSupportedCommand(SCSITarget *p) : inherited(p) { }
	virtual ~SCSICmdNotSupportedCommand() override { }

	SCSI::XferPhase Command(std::vector<uint8>& cmdseq) override {
		// XXX センスキーよく分からんけどとりあえず
		parent->sensekey = SCSI::SenseKey::IllegalRequest;
		parent->senseasc = SCSI::ASC::InvalidCommandOperationCode;
		status_byte = SCSI::StatusByte::CheckCondition;
		return SCSI::XferPhase::Status;
	}
};

#if defined(__clang__)
#pragma clang diagnostic pop
#endif

//
// SCSI ターゲットデバイスの共通部分
//

// コンストラクタ
SCSITarget::SCSITarget(const std::string& objname_,
	SCSIHostDevice *h, int id, std::vector<uint8>& inq)
	: inherited(objname_)
{
	inquiry_data = inq;
	// ホストへのリンク (所有関係のないポインタ)
	host = h;
	myid = id;
}

// デストラクタ
SCSITarget::~SCSITarget()
{
}

// 電源オン
bool
SCSITarget::PowerOn()
{
	ClearSense();
	cmdseq.clear();
	cmdlen = 0;
	cmd.reset();
	bytecount = 0;
	lastbyte = false;

	return true;
}

// バスリセットされた (SCSIBus から呼ばれる)
void
SCSITarget::BusResetCB()
{
	// リセットされたので持ってる信号を全部解放。
	// 一応自分が変えていい信号線だけを落とす。
	switch (GetPhase()) {
	 case SCSI::Phase::BusFree:
		break;
	 case SCSI::Phase::Arbitration:
	 case SCSI::Phase::Selection:
	 case SCSI::Phase::Reselection:
		NegateSEL();
		NegateBSY();
		break;
	 case SCSI::Phase::Transfer:
		NegateBSY();
		break;
	}

	// センスキーをクリア
	ClearSense();
}

// セレクションで選択された (SCSIBus から呼ばれる)
void
SCSITarget::SelectionCB()
{
	// ターゲットが、自身が選択されていることを認識してから BSY を
	// 立てるまでの時間はもうバス側でカウントしてあるので、ここでは
	// ただちに BSY を上げてよい。
	putlog(4, "Selection Callback");
	AssertBSY();
}

// 情報転送フェーズを開始する (SCSIBus から呼ばれる)
void
SCSITarget::StartTransferCB()
{
	// 最初はコマンドフェーズ。(ATN はとりあえず放置)
	putlog(4, "Start Command Phase");
	cmdseq.clear();
	cmdlen = 0;
	SetXfer(SCSI::XferPhase::Command);
	AssertREQ();
}

// 情報転送フェーズ間の遷移について
//
// 情報転送フェーズは TransferCB() で1バイトずつ転送を行い、所定のバイト数
// の転送が完了すればフェーズを遷移する。遷移方法は入出力方向別で6通り。
//  o IN  -> IN
//  o IN  -> OUT
//  o IN  -> END
//  o OUT -> OUT
//  o OUT -> IN
//  o OUT -> END
//
// IN -> IN の遷移。TransferCB() で前フェーズの最終バイト受信後、フェーズを
// 変更して REQ を立てるところまで。
//
// IN -> OUT の遷移。TransferCB() で前フェーズの最終バイト受信後、フェーズを
// 変更して (おそらく {Phase}Begin() を発行してから)、データをセットし REQ
// を立てる?
//
// IN -> END の遷移。TransferCB() で前フェーズの最終バイト受信後、BSY を
// 下げて終了。
//
// OUT -> OUT の遷移。前フェーズの最終バイトをイニシエータが受信後に
// TransferCB() が呼ばれるので、フェーズを変更して (おそらく {Phase}Begin()
// を発行してから)、データをセットし REQ を立てる?
//
// OUT -> IN の遷移。前フェーズの最終バイトをイニシエータが受信後に
// TransferCB() が呼ばれるので、フェーズを変更して (おそらく {Phase}Begin()
// を発行してから)、REQ を立てる?
//
// OUT -> END の遷移。前フェーズの最終バイトをイニシエータが受信後に
// TransferCB() が呼ばれるので、BSY を下げる?

// 情報転送フェーズでイニシエータが ACK を上げた。(SCSIBus から呼ばれる)
// OUT (イニシエータ → ターゲット方向) なら、ターゲットが上げた REQ に応答
// してイニシエータがデータをデータバスに乗せてからこれをコールしてくる。
// IN (ターゲット → イニシエータ方向) なら、ターゲットがデータをデータバスに
// 乗せて REQ を上げたのをイニシエータが受信した後でこれをコールしてくる。
void
SCSITarget::TransferCB()
{
	uint8 data;
	SCSI::XferPhase xfer;

	// このフェーズの最終バイトか
	lastbyte = false;

	// このフェーズの1バイト処理
	xfer = GetXfer();
	if (xfer == SCSI::XferPhase::Command) {
		// コマンドフェーズだけいろいろ特殊なので別処理。
		// 受信バッファが異なる、1バイト目で受信長が決まるなど。
		data = GetData();
		putlog(4, "TransferCB: cmdseq[%02d] = $%02x",
			(int)cmdseq.size(), data);
		cmdseq.push_back(data);
		if (cmdseq.size() == 1) {
			// 1バイト目ならこのコマンドが何バイトか調べる
			switch (cmdseq[0] >> 5) {
			 case 0:	cmdlen =  6;	break;
			 case 1:	cmdlen = 10;	break;
			 case 2:	cmdlen = 10;	break;
			 case 5:	cmdlen = 12;	break;
			 default:
				// 3,4 はリサーブ、6,7 はベンダー定義。XXX 来たらどうする?
				cmdlen = 6;
				break;
			}
		} else if (cmdseq.size() >= cmdlen) {
			if (loglevel >= 1) {
				std::string cmdbuf;
				const char *name = SCSI::GetCommandName(cmdseq[0]);
				cmdbuf += string_format("Command \"%s\"", name ?: "?");
				for (const auto& v : cmdseq) {
					cmdbuf += string_format(" $%02x", v);
				}
				putlogn("%s", cmdbuf.c_str());
			}
			// 最終バイトを受信したらコマンドを選択
			DispatchCmd();
			lastbyte = true;
		}
	} else if (GetIO() == false) {
		// OUT 方向 (ターゲットが受信側)
		// bytecount は減算方向。
		cmd->buf.push_back(GetData());
		bytecount--;
		if (bytecount == 0) {
			lastbyte = true;
		}
	} else {
		// IN 方向 (ターゲットが送信側)
		// bytecount は加算方向。
		// データは前のループの終わり (TransferAckCB()の下のほう) で
		// セットしているのでここでは引き取られた分をカウントする。
		bytecount++;
		if (bytecount >= cmd->buf.size()) {
			lastbyte = true;
		}
	}

	NegateREQ();
}

// 情報転送フェーズでイニシエータが ACK を下げた。(SCSIBus から呼ばれる)
void
SCSITarget::TransferAckCB()
{
	SCSI::XferPhase xfer = GetXfer();

	// これが最終バイトだったら、フェーズ完了 → フェーズ切り替え。
	if (lastbyte) {
		SCSI::XferPhase next;
		// フェーズ完了
		switch (xfer) {
		 case SCSI::XferPhase::Command:
			next = cmd->Command(cmdseq);
			break;
		 case SCSI::XferPhase::DataOut:
			next = cmd->DoneDataOut();
			break;
		 case SCSI::XferPhase::DataIn:
			next = cmd->DoneDataIn();
			break;
		 case SCSI::XferPhase::MsgOut:
			next = cmd->DoneMsgOut();
			break;
		 case SCSI::XferPhase::MsgIn:
			next = cmd->DoneMsgIn();
			break;
		 case SCSI::XferPhase::Status:
			next = cmd->DoneStatus();
			break;
		 default:
			PANIC("TransferCB unknown xfer done %d", (int)xfer);
		}
		putlog(2, "%s done", SCSI::GetXferPhaseName(xfer));

		// 簡略化のため、フェーズ完了時だけ REQ を引き伸ばすことができる
		if (cmd->optime != 0) {
			bus->SetOneshotReqWait(cmd->optime);
			cmd->optime = 0;
		}

		if (next != xfer) {
			// End ならバスフリーにして終了。
			if (next == SCSI::XferPhase::End) {
				NegateBSY();
				return;
			}

			// (必要なら)フェーズを更新
			SetXfer(next);
			xfer = next;

			// 新しいフェーズの開始処理
			// XXX ここで bytecount とかをリセットできればいいのだが
			// Command フェーズで Data フェーズの準備をすることが多いので
			// 悩ましい。
			switch (xfer) {
			 case SCSI::XferPhase::DataOut:
				cmd->buf.clear();
				bytecount = cmd->recvbytes;
				putlog(3, "DataOut begin (%d bytes)", bytecount);
				break;
			 case SCSI::XferPhase::DataIn:
				bytecount = 0;
				putlog(3, "DataIn begin (%d bytes)", (int)cmd->buf.size());
				break;
			 case SCSI::XferPhase::MsgOut:
				cmd->BeginMsgOut();
				cmd->buf.clear();
				bytecount = cmd->recvbytes;
				putlog(3, "MsgOut begin (%d bytes)", bytecount);
				break;
			 case SCSI::XferPhase::MsgIn:
				cmd->BeginMsgIn();
				bytecount = 0;
				putlog(3, "MsgIn begin (%d bytes)", (int)cmd->buf.size());
				break;
			 case SCSI::XferPhase::Status:
				cmd->BeginStatus();
				bytecount = 0;
				putlog(3, "Status begin (%d bytes)", (int)cmd->buf.size());
				break;
			 default:
				PANIC("TransferCB unknown xfer begin %d", (int)xfer);
			}
		}
	}

	// IN 方向 (ターゲットが送信側) ならここでデータをバスにセット。
	if (GetIO()) {
		SetData(cmd->buf[bytecount]);
		putlog(4, "SetData $%02x", GetData());
	}

	// 次のバイト転送を行うために REQ を上げる。
	AssertREQ();
}

// SCSI コマンドのディスパッチャ
void
SCSITarget::DispatchCmd()
{
	switch (cmdseq[0]) {
	 case SCSI::Command::TestUnitReady:	// 0x00
		cmd.reset(new SCSICmdTestUnitReady(this));
		break;

	 case SCSI::Command::RezeroUnit:	// 0x01
		cmd.reset(new SCSICmdRezeroUnit(this));
		break;

	 case SCSI::Command::RequestSense:	// 0x03
		cmd.reset(new SCSICmdRequestSense(this));
		break;

	 case SCSI::Command::Inquiry:		// 0x12
		cmd.reset(new SCSICmdInquiry(this));
		break;

	 case SCSI::Command::ModeSelect6:	// 0x15
		cmd.reset(new SCSICmdModeSelect6(this));
		break;

	 case SCSI::Command::ModeSense6:	// 0x1a
		cmd.reset(new SCSICmdModeSense6(this));
		break;

	 case SCSI::Command::ReadCapacity:	// 0x25
		cmd.reset(new SCSICmdTestUnitReady(this));
		break;

	 default:
		// サポートしていないコマンド
		const char *cname = SCSI::GetCommandName(cmdseq[0]);
		std::string name;
		if (cname) {
			name = string_format(" %s", cname);
		}
		putlog(0, "未実装 SCSI コマンド $%02x%s len=%d",
			cmdseq[0], name.c_str(), (int)cmdseq.size());

		cmd.reset(new SCSICmdNotSupportedCommand(this));
		break;
	}
}

//
// SCSI HD
//

// コンストラクタ
SCSIHD::SCSIHD(SCSIHostDevice *h, int id)
	: inherited(string_format("SCSIHD%d", id), h, id, inquiry_data)
{
	AddAlias(string_format("HD%d", id));

	devtype = SCSI::DevType::HD;
}

// デストラクタ
SCSIHD::~SCSIHD()
{
}

// 初期化
bool
SCSIHD::Init()
{
	const std::string myname = string_format("%s:ID%d",
		host->GetConfigName(), myid);
	const std::string keybody = string_format("%s-id%d-",
		host->GetConfigName(), myid);
	const std::string imgkey = keybody + "image";
	const std::string wpkey  = keybody + "writeprotect";
	const std::string stkey  = keybody + "seektime";

	const ConfigItem& itemimg = gConfig->Find(imgkey);
	const std::string& image = itemimg.AsString();
	// ここに来る時にチェックしているので ',' はあるはず
	auto pos = image.find(',') + 1;
	std::string filename = string_ltrim(image.substr(pos, image.size() - pos));
	if (filename.length() == 0) {
		itemimg.Err("imagefile not specified");
		return false;
	}

	// ファイルをオープン
	const std::string path = gMainApp.GetVMDir() + filename;
	file.SetFilename(path);
	file.SetDispname(string_format("%s \"%s\"", myname.c_str(), path.c_str()));
	mem = file.OpenFile();
	if (mem == NULL) {
		return false;
	}

	writeprotect = gConfig->Find(wpkey).AsInt();
	if (writeprotect) {
		putmsg(0, "%s: write protected", myname.c_str());
	}

	// 設定は [msec]、変数は [nsec]。
	seektime = gConfig->Find(stkey).AsInt();
	seektime *= 1_msec;

	return true;
}

// SCSI コマンドのディスパッチャ
void
SCSIHD::DispatchCmd()
{
	switch (cmdseq[0]) {
	 case SCSI::Command::Read6:			// 0x08
		cmd.reset(new SCSICmdRead6(this));
		break;

	 case SCSI::Command::Write6:		// 0x0a
		cmd.reset(new SCSICmdWrite6(this));
		break;

	 case SCSI::Command::StartStopUnit:	// 0x1b
		cmd.reset(new SCSICmdStartStopUnit(this));
		break;

	 case SCSI::Command::ReadCapacity:	// 0x25
		cmd.reset(new SCSICmdReadCapacity(this));
		break;

	 case SCSI::Command::Read10:		// 0x28
		cmd.reset(new SCSICmdRead10(this));
		break;

	 case SCSI::Command::Write10:		// 0x2a
		cmd.reset(new SCSICmdWrite10(this));
		break;

	 default:
		// 見付からなければ基本クラス側に任せる
		inherited::DispatchCmd();
		break;
	}
}

// 指定の場所を自由に読み出す
// (エミュレータ特権アクセス)
bool
SCSIHD::PeekImage(void *buf, uint32 start, uint32 len) const
{
	if (mem == NULL) {
		return false;
	}
	if ((uint64)start + (uint64)len > GetSize()) {
		return false;
	}
	memcpy(buf, mem + start, len);
	return true;
}

// HD の inquiry データ
/*static*/ std::vector<uint8>
SCSIHD::inquiry_data = {
	0x00,	// 00 Qualifier + DeviceType
	0x00,	// 01 RMB
	0x40,	// 02 ISO version
	0x00,	// 03 Response Data Format?
	36 - 4,	// 04 追加データ長
	0,		// 05 Reserved
	0,		// 06 Reserved
	0x00,	// 07
	'N', 'O', 'N', 'O', ' ', ' ', ' ', ' ',
	'S', 'C', 'S', 'I', 'H', 'D', ' ', ' ',
	' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
	'0', ' ', ' ', ' ',
};
