//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "sysport.h"
#include "mfp.h"
#include "ram.h"
#include "romimg_x68k.h"
#include "sram.h"

std::unique_ptr<SysportDevice> gSysport;

SysportDevice::SysportDevice()
	: inherited("SystemPort")
{
	AddAlias("SysPort");

	devaddr = baseaddr;
	devlen  = 0x2000;
}

SysportDevice::~SysportDevice()
{
}

void
SysportDevice::ResetHard()
{
	sysport.contrast = 0;	// XXX 未調査
	sysport.ram_wait = 0;
	sysport.rom_wait = 0;
	sysport.key_ctrl = false;
}

uint64
SysportDevice::Read(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case SYSPORT::CONTRAST:
		data = 0xf0 | sysport.contrast;
		break;
	 case SYSPORT::SCOPE3D:
		data = 0xff;
		break;
	 case SYSPORT::IMAGEUNIT:
		data = 0xff;
		break;
	 case SYSPORT::KEY:
		// キーボードは接続されてることにする
		data = 0x08;
		break;
	 case SYSPORT::WAIT:
		data = 0xff;
		break;
	 case SYSPORT::MPU:
		data = 0xdc;
		break;
	 case SYSPORT::SRAMWP:
		data = 0xff;
		break;
	 case SYSPORT::POWEROFF:
		data = 0xff;
		break;
	 default:
		__unreachable();
	}
	putlog(1, "$%06X -> $%02x", gMPU->GetPaddr(), data);
	return data;
}

uint64
SysportDevice::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case SYSPORT::CONTRAST:
		sysport.contrast = data & 15;
		break;

	 case SYSPORT::SCOPE3D:
	 case SYSPORT::IMAGEUNIT:
		break;

	 case SYSPORT::KEY:
		data &= 0x0e;
		// bit1,2
		if ((data & 0x06)) {
			putlog(0, "$E8E007 未サポート書き込み $%02x", data);
		}
		sysport.key_ctrl = (data & 0x08);
		break;

	 case SYSPORT::WAIT:
	 {
		// 設定値で保持
		sysport.rom_wait = data >> 4;
		sysport.ram_wait = data;

		// ROM に指示。常に設定値 + 2 でよい
		gIPLROM1->SetWait(sysport.rom_wait + 2);
		gIPLROM2->SetWait(sysport.rom_wait + 2);
		gCGROM->SetWait(sysport.rom_wait + 2);
		// RAM に指示。InsideOut p.124
		int wait = sysport.ram_wait;
		if (wait > 0)
			wait += 2;
		gRAM->SetWait(wait);
		break;
	 }

	 case SYSPORT::MPU:
		break;

	 case SYSPORT::SRAMWP:
		if (data == 0x31) {
			gSRAM->write_enable(true);
		} else {
			gSRAM->write_enable(false);
		}
		break;

	 case SYSPORT::POWEROFF:
		putlog(0, "$e8e00f (POWEROFF) 未サポート書き込み %02x", data);
		break;

	 default:
		__unreachable();
	}
	return 0;
}

uint64
SysportDevice::Peek(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case SYSPORT::CONTRAST:
		data = 0xf0 | sysport.contrast;
		break;
	 case SYSPORT::SCOPE3D:
		data = 0xff;
		break;
	 case SYSPORT::IMAGEUNIT:
		data = 0xff;
		break;
	 case SYSPORT::KEY:
		data = 0xff;
		break;
	 case SYSPORT::WAIT:
		data = 0xff;
		break;
	 case SYSPORT::MPU:
		data = 0xdc;
		break;
	 case SYSPORT::SRAMWP:
		data = 0xff;
		break;
	 case SYSPORT::POWEROFF:
		data = 0xff;
		break;
	 default:
		__unreachable();
	}
	return data;
}
