//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "zram.h"

//
// 3ポート共有 RAM
//
// ZRAM はロングワード単位でホストバイトオーダ配置なので、
// バイトアクセス、ワードアクセス時は HLB(), HLW() マクロ使用のこと。
//

std::unique_ptr<ZRAMDevice> gZRAM;

ZRAMDevice::ZRAMDevice()
	: inherited("ZRAM")
{
	devaddr = 0x71000000;
	devlen = 128 * 1024;

	ram.reset(new uint8[devlen]);
}

ZRAMDevice::~ZRAMDevice()
{
}

// アドレスデコーダ
inline uint64
ZRAMDevice::Decoder(uint32 addr) const
{
	// たぶん 128KB で折り返しイメージが見えるはず。要確認。
	return addr % devlen;
}

uint64
ZRAMDevice::Read8(uint32 addr)
{
	std::unique_lock<std::mutex> lock(mtx);

	uint32 offset = Decoder(addr);
	uint32 data = ram[HLB(offset)];
	return data;
}

uint64
ZRAMDevice::Read16(uint32 addr)
{
	std::unique_lock<std::mutex> lock(mtx);

	uint32 offset = Decoder(addr);
	uint32 data = *(uint16 *)&ram[HLW(offset)];
	return data;
}

uint64
ZRAMDevice::Read32(uint32 addr)
{
	std::unique_lock<std::mutex> lock(mtx);

	uint32 offset = Decoder(addr);
	uint32 data = *(uint32 *)&ram[offset];
	return data;
}

uint64
ZRAMDevice::Write8(uint32 addr, uint32 data)
{
	std::unique_lock<std::mutex> lock(mtx);

	uint32 offset = Decoder(addr);
	ram[HLB(offset)] = data;
	return 0;
}

uint64
ZRAMDevice::Write16(uint32 addr, uint32 data)
{
	std::unique_lock<std::mutex> lock(mtx);

	uint32 offset = Decoder(addr);
	*(uint16 *)&ram[HLW(offset)] = data;
	return 0;
}

uint64
ZRAMDevice::Write32(uint32 addr, uint32 data)
{
	std::unique_lock<std::mutex> lock(mtx);

	uint32 offset = Decoder(addr);
	*(uint32 *)&ram[offset] = data;
	return 0;
}

uint64
ZRAMDevice::Peek8(uint32 addr)
{
	std::unique_lock<std::mutex> lock(mtx);

	uint32 offset = Decoder(addr);
	return ram[HLB(offset)];
}
