//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "wxnono.h"
#include "monitor.h"
#include "vm.h"
#include <array>

enum {
	// ID は 0 から (wxID_LOWEST = 5000 未満まで) 使えるっぽい。
	// メニュー ID はメニュー ID 内で一意という制約があるだけ。
	// メニュー ID のうちウィンドウのオープン状態を維持する必要があるものは
	// (ID が 0 から使えるのなら) 0 から割り当てて配列で管理してしまうのが
	// 手っ取り早い。
	// ウィンドウに割り当てるウィンドウ ID とメニューイベントで飛んでくる
	// メニュー ID は互いに干渉しないはずなので、同じ値を使うことでメニューと
	// ウィンドウの紐付けをする。
	//
	// ウィンドウ(サブウィンドウ)は細かくは以下のように4つに分類できるが、
	// 処理上区別が必要なのは A. (デバッガの show コマンドとも共用できる
	// サブウィンドウ) とそれ以外のサブウィンドウだけ。
	//
	// -----------+------------+----
	// モニター   | テキスト   | A. 単純なテキストモニタウィンドウ
	// ウィンドウ | モニター   |    (SPC, MFP などのデバイスのモニター)
	//            | ウィンドウ +----
	//            |            | B. それ以外のテキストモニターウィンドウ
	//            |            |    (ログ、メモリのように表示位置があるもの、
	//            |            |     CMMU のように構造があるもの)
	//            +------------+----
	//            | それ以外の | C. ビデオモニターウィンドウなど
	//            | モニター   |    (ビットマッププレーン、TVRAM など)
	//            | ウィンドウ |
	// -----------+------------+----
	// それ以外のウィンドウ    | D. ソフトウェアキーボードウィンドウなど
	// ------------------------+----

	ID_SUBWIN_END_DUMMY = ID_SUBWIN_END,

	// ここから (ウィンドウでない)メニュー ID
	ID_FULL_SPEED,
	ID_RESTART_VM,
	ID_SCALE_START,
	ID_SCALE_1_00,
	ID_SCALE_0_75,
	ID_SCALE_0_50,
	ID_SCALE_END,

	ID_FONTSIZE_START, // START 以降は fontsize_t と同じ順に並べること
	ID_FONTSIZE_12,
	ID_FONTSIZE_16,
	ID_FONTSIZE_END,

	ID_MONRATE_START,
	ID_MONRATE_20,
	ID_MONRATE_5,
	ID_MONRATE_1,
	ID_MONRATE_END,

	ID_SHOW_STATPANEL,
	ID_MOUSEMODE,
	ID_RESET_BUTTON,
};

class WXMainFrame : public wxFrame
{
	using inherited = wxFrame;
 public:
	WXMainFrame(wxWindow *);
	virtual ~WXMainFrame() override;

	// ウィンドウをリストから削除する。
	// ウィンドウをクローズする際に呼び出すこと。
	void DeleteWindow(wxWindow *);

	// VM からの電源オフコールバック
	void PowerOffCallback();

	// VM からのダブルバスフォールト・コールバック
	void HaltCallback();

	// 起動時にオープンするサブウィンドウの一覧
	static std::array<bool, ID_SUBWIN_MAX> subwindow_start;

	// モニタウィンドウの更新頻度取得
	int GetMonitorRate() const { return monitor_rate; }

 private:
	void OnPaint(wxPaintEvent& event);
	void OnClose(wxCloseEvent& event);
	void OnActivate(wxActivateEvent& event);

	void OnFullSpeed(wxCommandEvent& event);
	void OnFullSpeedUI(wxUpdateUIEvent& event);
	void OnRestartVM(wxCommandEvent& event);

	void OnExit(wxCommandEvent&);

	void OnMenuScale(wxCommandEvent& event);
	void OnMenuScaleUI(wxUpdateUIEvent& event);
	void OnMenuFontSize(wxCommandEvent& event);
	void OnMenuFontSizeUI(wxUpdateUIEvent& event);
	void OnMenuMonRate(wxCommandEvent& event);
	void OnMenuMonRateUI(wxUpdateUIEvent& event);
	void OnShowStatusPanel(wxCommandEvent& event);
	void OnShowStatusPanelUI(wxUpdateUIEvent& event);

	void OnMouseMode(wxCommandEvent& event);
	void OnMouseModeUI(wxUpdateUIEvent& event);
	void OnResetButton(wxCommandEvent& event);

	void OnWindow(wxCommandEvent&);
	void DoWindow(int id, bool enable);
	void OnWindowUI(wxUpdateUIEvent&);

	void OnAbout(wxCommandEvent&);

	// ウィンドウをリストに追加する。
	void AddWindow(int id, wxWindow *);

	// 指定 ID のサブウィンドウがリストにあれば true を返す
	bool IsExistWindow(int id);

	// サブウィンドウリスト
	std::unique_ptr<wxWindow> windows[ID_SUBWIN_MAX] {};

	// モニタの更新頻度(これが実行中の一次情報)
	int monitor_rate {};
	// モニタ更新頻度 ID -> 数値変換
	static const int IDtoMonRate[];

	// ダブルバスフォールト・イベント
	void OnHaltNotify(wxCommandEvent& event);

	wxDECLARE_EVENT_TABLE();
};

extern WXMainFrame *gMainFrame;

// コールバック関数
extern void wxmainframe_poweroff_callback();
extern void wxmainframe_halt_callback(void *);
