//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "wxbitmapbuf.h"
#include "wxsubwindow.h"
#include "wxtextpanel.h"
#include "keyboard.h"
#include "vm.h"
#include <array>
#include <vector>

// キー描画用のデータ構造体
struct keydata_t
{
	// フォントサイズの切り替えに対応するため、座標系の単位はピクセルでは
	// ない点に注意。フォントサイズ(高さ)の 1/4 を 1 と表現する。つまり
	// どのフォントサイズでも半角1文字は 2 x 4。
	int x;	// このキー枠の左上点
	int y;	// このキー枠の左上点
	int w;	// このキー枠の幅
	int h;	// このキー枠の高さ

	// 色種別
	// LUNA の場合はキートップ色の区別に使う。1=濃い, 0=通常。
	// X68k では使用しない。
	int coltype;

	int led;			// -1ならLEDなし、0以上なら LED 番号
	uint keycode;		// キーコード (keyboard.h の KC_*)

	// 刻印
	// [0]: 通常
	// [1]: SHIFT時 (NULL なら通常のまま)
	const char *disp[2];

	int right() const { return x + w; }		// このキー枠の右
	int bottom() const { return y + h; }	// このキー枠の下
};

// ソフトウェアキーボード (パネル) 共通部分
class WXSoftKeyPanel : public WXTextPanel
{
	using inherited = WXTextPanel;
 protected:
	// 色の集合(パレット)
	struct ColorPalette {
		wxColour Background;	// 背景というかガワの部分の色
		wxColour Text;			// 文字色
		wxColour Keytop[2];		// キートップ色。0:通常色、1:(あれば)濃色
		wxColour KeyPressed;	// キーが押された時のキートップ色
		wxColour Border;		// LED 点灯時の枠の色
		wxColour LED[2];		// LED。色は各機種定義 (今のところ0:赤 1:緑)
	};

	// 修飾キーの状態
	enum Modifier {
		ModNormal = 0,
		ModShift,
	};

 protected:
	WXSoftKeyPanel(wxWindow *parent, const std::vector<keydata_t>& keydata_,
		int u_width_, int u_height_);
 public:
	virtual ~WXSoftKeyPanel() override;

	// フォントサイズ変更
	void SetFontSize(FontId fontid) override;

 protected:
	void InitLED();
	void DoSize();
	void OnPaint(wxPaintEvent& event);
	void OnTimer(wxTimerEvent& event);

	void Draw(wxDC& dc);
	void DrawKey(wxDC& dc, const keydata_t& key);
	void DrawEnterKey(wxDC& dc);
	void DrawEnterMark(wxDC& dc, const wxPoint& basep);
	virtual int DrawLED(wxDC& dc, const keydata_t& key) = 0;
	const keydata_t& FindKeydata(uint) const;

	void OnLeftDown(wxMouseEvent& event);
	void DoLeftDown(const keydata_t *key);
	void OnLeftUp(wxMouseEvent& event);
	void DoLeftUp(const keydata_t *key);
	void OnLeftDClick(wxMouseEvent& event);
	void OnRightDown(wxMouseEvent& event);
	bool IsModifier(uint code) const;
	void UpdateModifier();
	void KeyDown(const keydata_t& key);
	void KeyUp(const keydata_t& key);

	const keydata_t *GetKeydataFromPoint(const wxPoint& pt) const;

	// 左クリック押下時のキー
	const keydata_t *leftdown_key {};

	// 修飾キーの状態 (刻印を変更するため)
	Modifier modifier {};

	// 前回の状態
	std::vector<bool> last_pressed {};
	Modifier last_modifier {};

	// フォントサイズに相対的な描画単位
	int unit {};

	// コントロールのサイズ [基準単位]
	int u_width {};
	int u_height {};

	// コントロール全域のビットマップ
	WXBitmapBuf bitmap {};

	// エンター記号の周囲の点集合
	std::vector<wxPoint> enter_points {};
	wxRegion enter_region {};

	// 色
	ColorPalette color {};

	// UTF-8 -> CP932 変換
	// Unicode の "\\" は Shift_JIS に対応する文字がないため、これを
	// Unicode から Shift_JIS に変換すると "" になり、これではキートップが
	// 刻印できなくて困る。
	// CP932 なら "\\" を変換しても "\\" ('\x5c') になってくれるので、
	// これを使う。この \x5c をバックスラッシュだと思うか円記号だと思うかは
	// こちらで決める。
	wxCSConv conv { "CP932" };

	// キー描画データ (派生クラス側から渡される)
	const std::vector<keydata_t>& keydata;

	// タイマー
	wxTimer timer {};

	// イベントテーブル
	wxDECLARE_EVENT_TABLE();
};

// LUNA ソフトウェアキーボード (パネル)
class WXLunaSoftKeyPanel : public WXSoftKeyPanel
{
	using inherited = WXSoftKeyPanel;
 public:
	WXLunaSoftKeyPanel(wxWindow *parent);
	virtual ~WXLunaSoftKeyPanel() override;

 private:
	int DrawLED(wxDC& dc, const keydata_t& key) override;

	// キー描画データ
	static std::vector<keydata_t> keydata;
};

// X68k ソフトウェアキーボード (パネル)
class WXX68kSoftKeyPanel : public WXSoftKeyPanel
{
	using inherited = WXSoftKeyPanel;
 public:
	WXX68kSoftKeyPanel(wxWindow *parent);
	virtual ~WXX68kSoftKeyPanel() override;

 private:
	int DrawLED(wxDC& dc, const keydata_t& key) override;

	// キー描画データ
	static std::vector<keydata_t> keydata;
};

// ソフトウェアキーボード (ウィンドウ)
class WXSoftKeyWindow : public WXSubWindow
{
	using inherited = WXSubWindow;
 public:
	WXSoftKeyWindow(wxWindow *parent, vmtype_t vmtype);
	virtual ~WXSoftKeyWindow() override;

 private:
	// キーボードパネル
	WXSoftKeyPanel *panel {};
};
