//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "wxtextscreen.h"
#include "wxcolor.h"
#include "wxmainframe.h"
#include "bitrev.h"
#include "cgrom.h"
#include "sjis.h"

//
// テキストスクリーンを表示するパネル
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXTextScreen, inherited)
	EVT_PAINT(WXTextScreen::OnPaint)
	EVT_SIZE(WXTextScreen::OnSize)
	EVT_TIMER(wxID_ANY, WXTextScreen::OnTimer)
wxEND_EVENT_TABLE()

// コンストラクタ
WXTextScreen::WXTextScreen(wxWindow *parent, Monitor& monitor_)
	: inherited(parent)
	, monitor(monitor_)
{
	// モニタサイズから初期サイズを決定
	nnSize ms = monitor.GetSize();
	Init(ms.width, ms.height);

	wxSize size = wxSize(ms.width * font_width, ms.height * font_height);
	SetClientSize(size);
	SetMinClientSize(size);

	timer.SetOwner(this);

	// 現在の設定値を適用
	SetRate(gMainFrame->GetMonitorRate());

	// 最初に一回描画を起こす
	wxTimerEvent dummy;
	AddPendingEvent(dummy);
}

// デストラクタ
WXTextScreen::~WXTextScreen()
{
}

// (再)初期化
void
WXTextScreen::Init(int col, int row)
{
	ts.Init(col, row);
	prevbuf.resize(col * row);
	std::fill(prevbuf.begin(), prevbuf.end(), 0xffff);
}

// 画面更新頻度を Hz で設定する。
void
WXTextScreen::SetRate(int hz)
{
	timer.Start(1000 / hz);
}

// フォントサイズ変更 (外部インタフェース)
void
WXTextScreen::SetFontSize(FontId new_fontid)
{
	// 同じなら何もしない
	if (fontid == new_fontid) {
		return;
	}

	// フォントを更新
	UpdateFont(new_fontid);

	// 再描画を強制する。
	// テキストの桁数行数は変わってないので再確保はしなくていい。
	std::fill(prevbuf.begin(), prevbuf.end(), 0xffff);

	// コントロールの大きさを変更
	wxSize size = wxSize(ts.GetCol() * font_width, ts.GetRow() * font_height);
	SetClientSize(size);
	SetMinClientSize(size);
}

// サイズ変更イベント
void
WXTextScreen::OnSize(wxSizeEvent& event)
{
	// Mac ではウィンドウ作成時に 0 以下の値が来ることがある
	// GTK では 1 で来ることがある
	const wxSize& size = event.GetSize();
	if (size.x <= 1 || size.y <= 1) {
		return;
	}

	// テキストサイズが変わった時だけテキストバッファを再構成。
	// (フォントサイズを変えると桁数×行数が変わらず画面サイズが変わる
	// ということは起きることに注意)
	int col = size.x / font_width;
	int row = size.y / font_height;
	if (col != ts.GetCol() || row != ts.GetRow()) {
		Init(col, row);
	}

	// ビットマップサイズが変わった時だけビットマップバッファを再構成。
	wxSize bmpsize = bitmap.GetSize();
	wxSize newsize = wxSize(col * font_width, row * font_height);
	if (bmpsize != newsize) {
		// ビットマップ作成
		bitmap.Create(newsize);
	}
}

// タイマーイベント
void
WXTextScreen::OnTimer(wxTimerEvent& event)
{
	DoRefresh();
}

// 画面を更新して再描画する。
//
// 通常はタイマーイベントにより更新間隔ごとに呼ばれるが、
// ユーザ操作により画面の更新が必要ならその都度直接これを呼ぶ。
void
WXTextScreen::DoRefresh()
{
	// モニタースクリーンを更新して
	MONITOR_UPDATE(monitor, ts);
	// 必要なら再描画
	if (ts.GetBuf() != prevbuf) {
		Refresh();
	}
}

// 描画イベント (UI スレッドから呼ばれる)
void
WXTextScreen::OnPaint(wxPaintEvent& event)
{
	// メモリ DC に描画
	wxMemoryDC memDC;
	memDC.SelectObject(bitmap);
	// 基本的には RefreshIfUpdated() により更新があった時だけ呼ばれる
	Draw(memDC);

	// 実画面 DC にコピー
	wxPaintDC dstDC(this);
	wxSize size = bitmap.GetSize();
	dstDC.Blit(0, 0, size.x, size.y, &memDC, 0, 0);
}

// テキストバッファの内容をパネルのバックグラウンドバッファに描画する。
void
WXTextScreen::Draw(wxDC& dc)
{
	auto text = ts.GetBuf().begin();
	auto prev = prevbuf.begin();

	int col = ts.GetCol();
	int row = ts.GetRow();
	int px = 0;
	int py = 0;
	uint prevattr = (uint)-1;

	for (int y = 0; y < row; y++, py += font_height) {
		px = 0;
		for (int x = 0; x < col; ) {
			uint16 data = *text;
			uint attr = (data & 0xff00);
			uint chr  = (data & 0xff);

			// 属性(色)を変更
			if (__predict_false(attr != prevattr)) {
				switch (attr) {
				 case TA::On:
					SetTextColor(BGPANEL, *wxBLACK);
					break;
				 case TA::Disable:
					SetTextColor(UD_GREY, BGPANEL);
					break;
				 default:
					ResetTextColor();
					break;
				}
				prevattr = attr;
			}

			if (__predict_true(SJIS::IsHankaku(chr))) {
				// 半角文字
				if (*prev != data) {
					DrawChar1(dc, px, py, chr, attr);
					if (SJIS::IsZenkaku(*prev) && x < col - 1) {
						// 変更前が全角文字だったら 2 バイト目も消す
						prev[1] = 0;
					}
					*prev = data;
				}
				prev += 1;
				text += 1;
				x += 1;
				px += font_width * 1;
			} else {
				// 全角文字
				// XXX: 最右カラムだったら?
				if (*prev != data || prev[1] != text[1]) {
					uint code = (chr << 8) | (text[1] & 0xff);
					DrawChar2(dc, px, py, code, attr);
					if (SJIS::IsASCII(*prev) && SJIS::IsZenkaku(prev[1])
					 && x < col - 2) {
						// 変更前が 'Aあ' だったら'あ'の2バイト目も消す
						prev[2] = 0;
					}
					prev[0] = data;
					prev[1] = text[1];
				}
				prev += 2;
				text += 2;
				x += 2;
				px += font_width * 2;
			}
		}
	}
}
